# Imports
import os
import threading
import time
from functools import partial
import json
import copy
import math
import webbrowser
import sys
from matplotlib.figure import Figure
import warnings 
from pathlib import Path
from tkinter import *
import tkinter as tk
from tkinter import ttk
from tkinter import filedialog, messagebox
from importlib.metadata import version
import traceback

# External libraries
import pandas as pd
import seaborn as sns
from scipy import stats
import numpy as np
from matplotlib.backends.backend_tkagg import (FigureCanvasTkAgg,  NavigationToolbar2Tk) 
import sv_ttk

# Import the MEA library
try:
    from .mea import *
except:
    from mea import *
    
def MEA_GUI():
    # Setting the warnings to be ignored 
    warnings.filterwarnings('ignore') 

    # Initialize tkinter GUI
    root = Tk()

    """Visual settings"""
    # Set the theme
    global theme
    theme='dark'
    sv_ttk.set_theme(theme)

    # Define font and styles
    font="Arial"
    def define_styles():
        # Define different styles
        labelframestyle = ttk.Style()
        labelframestyle.configure("Custom.TLabelframe.Label", font=(font, 20))

        subframestyle = ttk.Style()
        subframestyle.configure("Sublabel.TLabelframe.Label", font=(font, 12))

        bigbuttonstyle=ttk.Style()
        bigbuttonstyle.configure("bigbutton.TButton", font=(font, 30), anchor="center", justify="center", wraplength=400)

        mediumbuttonstyle=ttk.Style()
        mediumbuttonstyle.configure("mediumbutton.TButton", font=(font, 15))

        wellbuttonstyle=ttk.Style()
        wellbuttonstyle.configure("wellbutton.TButton", font=(font, 15))

    define_styles()

    # Add function for changing theme    
    def toggle_theme():
        sv_ttk.toggle_theme()
        global theme
        theme=sv_ttk.get_theme()
        # For some reason you have to redefine the styles everytime you switch themes
        define_styles()

    # Set correct icons
    def resource_path(relative_path):
        """ Get absolute path to resource, works for dev and for PyInstaller """
        try:
            # PyInstaller creates a temp folder and stores path in _MEIPASS
            base_path = sys._MEIPASS
        except Exception:
            source_path = Path(__file__).resolve()
            base_path = source_path.parent
        return os.path.join(base_path, relative_path)
    application_path=resource_path("cureq_icon.ico")

    try:
        root.iconbitmap(application_path)
    except Exception as error:
        print("Could not load in icon")
        print(error)


    root.title(f"CureQ MEA analysis tool - Version: {version('CureQ')}")

    # Define global variables and default values
    global filename
    global validation_method
    filename=""
    validation_method='DMP_noisebased'

    def go_to_parameterframe():
        parameterframe.pack(fill='both', expand=True)
        main_frame.pack_forget()

    def openfiles():
        global filename
        selectedfile = filedialog.askopenfilename(filetypes=[("MEA data", "*.h5")])
        if len(selectedfile)!=0:
            filename=selectedfile
            selectedfile=os.path.split(filename)[1]
            btn_selectfile.configure(text=selectedfile)
            # selectedfilelabel.config(text=selectedfile)

    class Tooltip:
        def __init__(self, widget, text):
            self.widget = widget
            self.text = text
            self.tooltip = None
            
        def show_tooltip(self, event):
            x, y, _, _ = self.widget.bbox("insert")
            x += self.widget.winfo_rootx() + 25
            y += self.widget.winfo_rooty() + 25
            
            # Create tooltip if not exists
            if not self.tooltip:
                self.tooltip = tk.Toplevel(self.widget)
                self.tooltip.wm_overrideredirect(True)
                self.tooltip.wm_geometry(f"+{x}+{y}")
                label = ttk.Label(self.tooltip, text=self.text, borderwidth=3, relief='sunken')
                label.pack()
                
        def hide_tooltip(self, event):
            if self.tooltip:
                self.tooltip.destroy()
                self.tooltip = None


    main_frame=ttk.Frame(root)
    main_frame.pack(fill='both', expand=True)

    # Scrollable frame
    class VerticalScrolledFrame(ttk.Frame):
        def __init__(self, parent, width=1600, height=1000, *args, **kw):
            # You can pass width and height to the constructor to set default size.
            ttk.Frame.__init__(self, parent, *args, **kw)

            # Create a canvas object and a vertical scrollbar for scrolling it.
            vscrollbar = ttk.Scrollbar(self, orient=VERTICAL)
            vscrollbar.pack(fill=Y, side=RIGHT, expand=FALSE)
            
            # Set the canvas size by passing width and height
            canvas = tk.Canvas(self, bd=0, highlightthickness=0, 
                            yscrollcommand=vscrollbar.set, width=width, height=height)
            canvas.pack(side=LEFT, fill=BOTH, expand=TRUE)
            vscrollbar.config(command=canvas.yview)

            # Reset the view
            canvas.xview_moveto(0)
            canvas.yview_moveto(0)

            # Create a frame inside the canvas which will be scrolled with it.
            self.interior = interior = ttk.Frame(canvas)
            interior_id = canvas.create_window(0, 0, window=interior, anchor=NW)

            # Track changes to the canvas and frame width and sync them,
            # also updating the scrollbar.
            def _configure_interior(event):
                # Update the scrollbars to match the size of the inner frame.
                size = (interior.winfo_reqwidth(), interior.winfo_reqheight())
                canvas.config(scrollregion="0 0 %s %s" % size)
                if interior.winfo_reqwidth() != canvas.winfo_width():
                    # Update the canvas's width to fit the inner frame.
                    canvas.config(width=interior.winfo_reqwidth())
            interior.bind('<Configure>', _configure_interior)

            def _configure_canvas(event):
                if interior.winfo_reqwidth() != canvas.winfo_width():
                    # Update the inner frame's width to fill the canvas.
                    canvas.itemconfigure(interior_id, width=canvas.winfo_width())
            canvas.bind('<Configure>', _configure_canvas)

            def _on_mousewheel(event):
                canvas.yview_scroll(int(-1*(event.delta/120)), "units")

            canvas.bind_all("<MouseWheel>", _on_mousewheel)

    # Frame for the sidebar buttons
    sidebarframe=ttk.Frame(master=main_frame)
    sidebarframe.grid(row=0, column=2, rowspan=10, sticky='nesw')

    # Switch between light and dark mode
    theme_switch = ttk.Checkbutton(sidebarframe, text="Light theme", style="Switch.TCheckbutton", command=toggle_theme)
    theme_switch.grid(row=0, column=0, sticky='nesw', pady=10, padx=10)

    def link_to_cureq():
        webbrowser.open_new("https://cureq.nl/")
    cureq_button=ttk.Button(master=sidebarframe, text="CureQ", command=link_to_cureq)
    cureq_button.grid(row=1, column=0, sticky='nesw', pady=10, padx=10)

    def link_to_pypi():
        webbrowser.open_new("https://pypi.org/project/CureQ/")
    pypi_button=ttk.Button(master=sidebarframe, text="Library", command=link_to_pypi)
    pypi_button.grid(row=2, column=0, sticky='nesw', pady=10, padx=10)

    def link_to_github():
        webbrowser.open_new("https://github.com/CureQ")
    github_button=ttk.Button(master=sidebarframe, text="Github", command=link_to_github)
    github_button.grid(row=3, column=0, sticky='nesw', pady=10, padx=10)

    # Set up the frame where te user selects different parameters
    parameterframe=VerticalScrolledFrame(root)
    parameterframe.pack(fill='both', expand=True)
    parameterframe.pack_forget()

    btn_parameter=ttk.Button(master=main_frame, text="Set\nParameters", compound=LEFT, style="bigbutton.TButton", command=go_to_parameterframe)
    btn_parameter.grid(row=0, column=0, padx=10, pady=10, sticky='nsew', ipadx=25, ipady=25)

    # Set up the required parameters
    requiredparameters=ttk.LabelFrame(parameterframe.interior, text='Required Parameters', style="Custom.TLabelframe")
    requiredparameters.grid(row=0, column=0, padx=10, pady=10, sticky='nsew')

    btn_selectfile=ttk.Button(master=requiredparameters, text="Choose a file", command=openfiles)
    btn_selectfile.grid(row=0, column=0, columnspan=2, padx=10, pady=10, sticky='nesw')

    # Setup the sampling frequency
    hertzlabel=ttk.Label(master=requiredparameters, text="Sampling rate:", font=(font,10))
    hertzlabel.grid(row=2, column=0, padx=10, pady=10, sticky='w')
    hertzinput=ttk.Entry(master=requiredparameters)
    hertzinput.grid(row=2, column=1, padx=10, pady=10, sticky='w')

    # Setup up the amount of electrodes
    electrodeamntlabel=ttk.Label(master=requiredparameters, text="Electrode amount:", font=(font,10))
    electrodeamntlabel.grid(row=3, column=0, padx=10, pady=10, sticky='w')
    electrodeamnttooltip = Tooltip(electrodeamntlabel, 'How many electrodes does each well contain?')
    electrodeamntlabel.bind("<Enter>", electrodeamnttooltip.show_tooltip)
    electrodeamntlabel.bind("<Leave>", electrodeamnttooltip.hide_tooltip)
    electrodeamntinput=ttk.Entry(master=requiredparameters)
    electrodeamntinput.grid(row=3, column=1, padx=10, pady=10, sticky='w')

    # Set up all the filter parameters
    filterparameters=ttk.LabelFrame(parameterframe.interior, text='Filter', style="Custom.TLabelframe")
    filterparameters.grid(row=0, column=1, padx=10, pady=10, sticky='nsew')

    # Low cutoff
    lowcutofflabel=ttk.Label(master=filterparameters, text="Low cutoff:", font=(font,10))
    lowcutofflabel.grid(row=1, column=0, padx=10, pady=10, sticky='w')
    lowcutofftooltip = Tooltip(lowcutofflabel, 'Define the low cutoff value for the butterworth bandpass filter. Values should be given in hertz')
    lowcutofflabel.bind("<Enter>", lowcutofftooltip.show_tooltip)
    lowcutofflabel.bind("<Leave>", lowcutofftooltip.hide_tooltip)
    lowcutoffinput=ttk.Entry(master=filterparameters)
    lowcutoffinput.grid(row=1, column=1, padx=10, pady=10, sticky='w')

    # High cutoff
    highcutofflabel=ttk.Label(master=filterparameters, text="High cutoff:", font=(font,10))
    highcutofflabel.grid(row=2, column=0, padx=10, pady=10, sticky='w')
    highcutofftooltip = Tooltip(highcutofflabel, 'Define the high cutoff value for the butterworth bandpass filter. Values should be given in hertz')
    highcutofflabel.bind("<Enter>", highcutofftooltip.show_tooltip)
    highcutofflabel.bind("<Leave>", highcutofftooltip.hide_tooltip)
    highcutoffinput=ttk.Entry(master=filterparameters)
    highcutoffinput.grid(row=2, column=1, padx=10, pady=10, sticky='w')

    # Filter order
    orderlabel=ttk.Label(master=filterparameters, text="Filter order:", font=(font,10))
    orderlabel.grid(row=3, column=0, padx=10, pady=10, sticky='w')
    ordertooltip = Tooltip(orderlabel, 'The filter order for the butterworth filter')
    orderlabel.bind("<Enter>", ordertooltip.show_tooltip)
    orderlabel.bind("<Leave>", ordertooltip.hide_tooltip)
    orderinput=ttk.Entry(master=filterparameters)
    orderinput.grid(row=3, column=1, padx=10, pady=10, sticky='w')

    # Set up all the spike detection parameters
    spikedetectionparameters=ttk.LabelFrame(parameterframe.interior, text='Spike Detection', style="Custom.TLabelframe")
    spikedetectionparameters.grid(row=0, column=2, padx=10, pady=10, sticky='nsew', rowspan=2)

    # Threshold parameters
    thresholdparameters=ttk.LabelFrame(spikedetectionparameters, text='Threshold parameters', style="Sublabel.TLabelframe")
    thresholdparameters.grid(row=0, column=0, padx=10, pady=10, sticky='nsew')

    # Threshold portion
    thresholdportionlabel=ttk.Label(master=thresholdparameters, text="Threshold portion:", font=(font,10))
    thresholdportionlabel.grid(row=1, column=0, padx=10, pady=10, sticky='w')
    thresholdportiontooltip = Tooltip(thresholdportionlabel, 'Define the portion of the electrode data that is used for determining the threshold.\nA higher values will give a better estimate of the background noise,\nbut will take longer to compute\nRanges from 0 to 1.')
    thresholdportionlabel.bind("<Enter>", thresholdportiontooltip.show_tooltip)
    thresholdportionlabel.bind("<Leave>", thresholdportiontooltip.hide_tooltip)
    thresholdportioninput=ttk.Entry(master=thresholdparameters)
    thresholdportioninput.grid(row=1, column=1, padx=10, pady=10, sticky='w')

    # stdevmultiplier
    stdevmultiplierlabel=ttk.Label(master=thresholdparameters, text="Standard Deviation Multiplier:", font=(font,10))
    stdevmultiplierlabel.grid(row=2, column=0, padx=10, pady=10, sticky='w')
    stdevmultipliertooltip = Tooltip(stdevmultiplierlabel, 'Define when beyond which point values are seen as\noutliers (spikes) when identifying spike-free noise\nA higher value will identify more data as noise')
    stdevmultiplierlabel.bind("<Enter>", stdevmultipliertooltip.show_tooltip)
    stdevmultiplierlabel.bind("<Leave>", stdevmultipliertooltip.hide_tooltip)
    stdevmultiplierinput=ttk.Entry(master=thresholdparameters)
    stdevmultiplierinput.grid(row=2, column=1, padx=10, pady=10, sticky='w')

    # RMSmultiplier
    RMSmultiplierlabel=ttk.Label(master=thresholdparameters, text="RMS Multiplier:", font=(font,10))
    RMSmultiplierlabel.grid(row=3, column=0, padx=10, pady=10, sticky='w')
    RMSmultipliertooltip = Tooltip(RMSmultiplierlabel, 'Define the multiplication factor of the root mean square (RMS) of the background noise\nA higher number will lead to a higher threshold')
    RMSmultiplierlabel.bind("<Enter>", RMSmultipliertooltip.show_tooltip)
    RMSmultiplierlabel.bind("<Leave>", RMSmultipliertooltip.hide_tooltip)
    RMSmultiplierinput=ttk.Entry(master=thresholdparameters)
    RMSmultiplierinput.grid(row=3, column=1, padx=10, pady=10, sticky='w')

    # Spike validation parameters
    validationparameters=ttk.LabelFrame(spikedetectionparameters, text='Spike Validation Parameters', style="Sublabel.TLabelframe")
    validationparameters.grid(row=1, column=0, padx=10, pady=10, sticky='nsew')

    # Refractory period
    refractoryperiodlabel=ttk.Label(master=validationparameters, text="Refractory Period:", font=(font,10))
    refractoryperiodlabel.grid(row=0, column=0, padx=10, pady=10, sticky='w')
    refractoryperiodtooltip = Tooltip(refractoryperiodlabel, 'Define the refractory period in the spike detection\nIn this period after a spike, no other spike can be detected\nValue should be given in seconds, so 2 ms = 0.002 s')
    refractoryperiodlabel.bind("<Enter>", refractoryperiodtooltip.show_tooltip)
    refractoryperiodlabel.bind("<Leave>", refractoryperiodtooltip.hide_tooltip)
    refractoryperiodinput=ttk.Entry(master=validationparameters)
    refractoryperiodinput.grid(row=0, column=1, padx=10, pady=10, sticky='w')

    # Dropdown menu where the user selects the validation method
    def option_selected(event):
        global validation_method
        validation_method = dropdown_var.get()
        if validation_method=='DMP_noisebased':
            exittimeinput.configure(state="enabled")
            maxheightinput.configure(state="enabled")
            amplitudedropinput.configure(state="enabled")
        else:
            exittimeinput.configure(state="disabled")
            maxheightinput.configure(state="disabled")
            amplitudedropinput.configure(state="disabled")
        
        
    dropdown_var = tk.StringVar(validationparameters)
    options = ['DMP_noisebased', 'none']
    dropdownlabel = ttk.Label(master=validationparameters, text="Spike validation method:", font=(font,10))
    dropdownlabel.grid(row=1, column=0, padx=10, pady=10, sticky='w')
    dropdown_menu = ttk.OptionMenu(validationparameters, dropdown_var, options[0], *options, command=option_selected)
    dropdown_menu.grid(row=1, column=1, padx=10, pady=10, sticky='nesw')

    exittimelabel=ttk.Label(master=validationparameters, text="Exit time:", font=(font,10))
    exittimelabel.grid(row=2, column=0, padx=10, pady=10, sticky='w')
    exittimetooltip = Tooltip(exittimelabel, 'Define the time a spike gets to drop a certain value.\nStarts from the peak of the spike.\nValue should be given in seconds, so 0.24 ms is 0.00024s')
    exittimelabel.bind("<Enter>", exittimetooltip.show_tooltip)
    exittimelabel.bind("<Leave>", exittimetooltip.hide_tooltip)
    exittimeinput=ttk.Entry(master=validationparameters)
    exittimeinput.grid(row=2, column=1, padx=10, pady=10, sticky='w')

    amplitudedroplabel=ttk.Label(master=validationparameters, text="Drop amplitude:", font=(font,10))
    amplitudedroplabel.grid(row=3, column=0, padx=10, pady=10, sticky='w')
    amplitudedroptooltip = Tooltip(amplitudedroplabel, 'Multiplied with the standard deviation of the surrounding noise.\nThis is the height the spike will have to drop in\na certain amount of time to be registered')
    amplitudedroplabel.bind("<Enter>", amplitudedroptooltip.show_tooltip)
    amplitudedroplabel.bind("<Leave>", amplitudedroptooltip.hide_tooltip)
    amplitudedropinput=ttk.Entry(master=validationparameters)
    amplitudedropinput.grid(row=3, column=1, padx=10, pady=10, sticky='w')

    maxheightlabel=ttk.Label(master=validationparameters, text="Max drop amount:", font=(font,10))
    maxheightlabel.grid(row=4, column=0, padx=10, pady=10, sticky='w')
    maxheighttooltip = Tooltip(maxheightlabel, 'Multiplied with the threshold.\nThe maximum height a spike can be required to drop in amplitude in the set timeframe')
    maxheightlabel.bind("<Enter>", maxheighttooltip.show_tooltip)
    maxheightlabel.bind("<Leave>", maxheighttooltip.hide_tooltip)
    maxheightinput=ttk.Entry(master=validationparameters)
    maxheightinput.grid(row=4, column=1, padx=10, pady=10, sticky='w')

    def parameter_to_main_func():
        # Check if the user has selected a file
        if filename == "":
            tk.messagebox.showerror(title='Error', message='No file has been selected yet, please select a file')
            return

        # save parameters in global values
        global wells
        wells='all'
        global hertz
        global low_cutoff
        global high_cutoff
        global order
        global spikeduration
        global exit_time_s
        global plot_electrodes
        plot_electrodes = False
        global electrode_amnt
        global kde_bandwidth
        global smallerneighbours
        global minspikes_burst
        global max_threshold
        global default_threshold
        global max_drop_amount
        global amplitude_drop_sd
        global stdevmultiplier
        global RMSmultiplier
        global min_channels
        global threshold_method
        global activity_threshold
        global threshold_portion
        global remove_inactive_electrodes
        global cut_data_bool
        global parts
        global global_nbd_kde_bandwidth
        # Save parameters
        try:
            hertz=int(hertzinput.get())
            low_cutoff=int(lowcutoffinput.get())
            high_cutoff=int(highcutoffinput.get())
            order=int(orderinput.get())
            spikeduration=float(refractoryperiodinput.get())
            exit_time_s=float(exittimeinput.get())
            electrode_amnt=int(electrodeamntinput.get())
            kde_bandwidth=float(isikdebwinput.get())
            smallerneighbours=int(smallernbinput.get())
            minspikes_burst=int(minspikesinput.get())
            max_threshold=float(maxisiinput.get())
            default_threshold=float(defaultthinput.get())
            max_drop_amount=float(maxheightinput.get())
            amplitude_drop_sd=float(amplitudedropinput.get())
            stdevmultiplier=float(stdevmultiplierinput.get())
            RMSmultiplier=float(RMSmultiplierinput.get())
            min_channels=float(minchannelsinput.get())
            threshold_method=networkth_var.get()
            global_nbd_kde_bandwidth=float(nbd_kde_bandwidth_input.get())
            remove_inactive_electrodes=bool(removeinactivevar.get())
            activity_threshold=float(activitythinput.get())
            threshold_portion=float(thresholdportioninput.get())
            cut_data_bool=bool(splitdatavar.get())
            parts=int(splitdatapartsinput.get())

            main_frame.pack(fill='both', expand=True)
            parameterframe.pack_forget()
        except Exception as error:
            traceback.print_exc()
            tk.messagebox.showerror(title='Error', message='Certain parameters could not be converted to the correct datatype (e.g. int or float). Please check if every parameter has the correct values')

    parameter_to_main=ttk.Button(master=parameterframe.interior, text="Save parameters and return", command=parameter_to_main_func)
    parameter_to_main.grid(row=3, column=0, padx=10, pady=(10,20), sticky='nsew')

    def parameter_to_main_func_no_save():
        main_frame.pack(fill='both', expand=True)
        parameterframe.pack_forget()

    parameter_to_main_no_save=ttk.Button(master=parameterframe.interior, text="Do not save parameters and return to main menu", command=parameter_to_main_func_no_save)
    parameter_to_main_no_save.grid(row=4, column=0, padx=10, pady=(0, 0), sticky='nsew', columnspan=3)

    # Set up all the burst detection parameters
    burstdetectionparameters=ttk.LabelFrame(parameterframe.interior, text='Burst Detection', style="Custom.TLabelframe")
    burstdetectionparameters.grid(row=1, column=0, padx=10, pady=10, sticky='nsew')

    # Setup up the minimal amount of spikes for a burst
    minspikeslabel=ttk.Label(master=burstdetectionparameters, text="Minimal amount of spikes:", font=(font,10))
    minspikeslabel.grid(row=0, column=0, padx=10, pady=10, sticky='w')
    minspikestooltip = Tooltip(minspikeslabel, 'Define the minimal amount of spikes a burst should have before being considered as one')
    minspikeslabel.bind("<Enter>", minspikestooltip.show_tooltip)
    minspikeslabel.bind("<Leave>", minspikestooltip.hide_tooltip)
    minspikesinput=ttk.Entry(master=burstdetectionparameters)
    minspikesinput.grid(row=0, column=1, padx=10, pady=10, sticky='w')

    # Setup up the default threshold
    defaultthlabel=ttk.Label(master=burstdetectionparameters, text="Default interval threshold:", font=(font,10))
    defaultthlabel.grid(row=1, column=0, padx=10, pady=10, sticky='w')
    defaultthtooltip = Tooltip(defaultthlabel, 'Define the default inter-spike interval threshold\nthat is used for burst detection\nValue should be given in miliseconds')
    defaultthlabel.bind("<Enter>", defaultthtooltip.show_tooltip)
    defaultthlabel.bind("<Leave>", defaultthtooltip.hide_tooltip)
    defaultthinput=ttk.Entry(master=burstdetectionparameters)
    defaultthinput.grid(row=1, column=1, padx=10, pady=10, sticky='w')

    # Setup up the max threshold
    maxisilabel=ttk.Label(master=burstdetectionparameters, text="Max interval threshold:", font=(font,10))
    maxisilabel.grid(row=2, column=0, padx=10, pady=10, sticky='w')
    maxisitooltip = Tooltip(maxisilabel, 'Define the maximum value the inter-spike interval threshold\ncan be when 2 peaks have been detected in the ISI graph\nValue should be given in miliseconds')
    maxisilabel.bind("<Enter>", maxisitooltip.show_tooltip)
    maxisilabel.bind("<Leave>", maxisitooltip.hide_tooltip)
    maxisiinput=ttk.Entry(master=burstdetectionparameters)
    maxisiinput.grid(row=2, column=1, padx=10, pady=10, sticky='w')

    # Setup the KDE bandwidth
    isikdebwlabel=ttk.Label(master=burstdetectionparameters, text="KDE bandwidth:", font=(font,10))
    isikdebwlabel.grid(row=3, column=0, padx=10, pady=10, sticky='w')
    isikdebwtooltip = Tooltip(isikdebwlabel, 'Define the bandwidth that is used when calculating the\nkernel density estimate of the inter-spike intervals')
    isikdebwlabel.bind("<Enter>", isikdebwtooltip.show_tooltip)
    isikdebwlabel.bind("<Leave>", isikdebwtooltip.hide_tooltip)
    isikdebwinput=ttk.Entry(master=burstdetectionparameters)
    isikdebwinput.grid(row=3, column=1, padx=10, pady=10, sticky='w')

    # Setup the amount of smaller neighbours
    smallernblabel=ttk.Label(master=burstdetectionparameters, text="Smaller neighbours:", font=(font,10))
    smallernblabel.grid(row=4, column=0, padx=10, pady=10, sticky='w')
    smallernbtooltip = Tooltip(smallernblabel, 'Define the amount of values next to the peak, that should be lower than the peak\nin order for it to be considered a peak in the distribution')
    smallernblabel.bind("<Enter>", smallernbtooltip.show_tooltip)
    smallernblabel.bind("<Leave>", smallernbtooltip.hide_tooltip)
    smallernbinput=ttk.Entry(master=burstdetectionparameters)
    smallernbinput.grid(row=4, column=1, padx=10, pady=10, sticky='w')

    # Set up all the network burst detection parameters
    networkburstdetectionparameters=ttk.LabelFrame(parameterframe.interior, text='Network Burst Detection', style="Custom.TLabelframe")
    networkburstdetectionparameters.grid(row=1, column=1, padx=10, pady=10, sticky='nsew')

    # Setup the minimum amount of channels participating
    minchannelslabel=ttk.Label(master=networkburstdetectionparameters, text="Min channels:", font=(font,10))
    minchannelslabel.grid(row=0, column=0, padx=10, pady=10, sticky='w')
    minchannelstooltip = Tooltip(minchannelslabel, 'Define the minimal percentage of channels that should be active in a network burst, values ranges from 0 to 1')
    minchannelslabel.bind("<Enter>", minchannelstooltip.show_tooltip)
    minchannelslabel.bind("<Leave>", minchannelstooltip.hide_tooltip)
    minchannelsinput=ttk.Entry(master=networkburstdetectionparameters)
    minchannelsinput.grid(row=0, column=1, padx=10, pady=10, sticky='w')

    # Setup the thresholding method
    networkth_var = tk.StringVar(networkburstdetectionparameters)
    nwthoptions = ['Yen', 'Otsu', 'Li', 'Isodata', 'Mean', 'Minimum', 'Triangle']
    dropdownlabel = ttk.Label(master=networkburstdetectionparameters, text="Thresholding method:", font=(font,10))
    dropdownlabel.grid(row=1, column=0, padx=10, pady=10, sticky='w')
    dropdown_menu = ttk.OptionMenu(networkburstdetectionparameters, networkth_var, nwthoptions[0], *nwthoptions)
    dropdown_menu.grid(row=1, column=1, padx=10, pady=10, sticky='nesw')

    # Setup the network burst detection KDE bandwidth
    nbd_kde_bandwidth_label=ttk.Label(master=networkburstdetectionparameters, text="KDE Bandwidth:", font=(font,10))
    nbd_kde_bandwidth_label.grid(row=2, column=0, padx=10, pady=10, sticky='w')
    nbd_kde_bandwidth_tooltip = Tooltip(nbd_kde_bandwidth_label, 'Define the bandwidth value that should be used when creating\nthe kernel density estimate for the network burst detection')
    nbd_kde_bandwidth_label.bind("<Enter>", nbd_kde_bandwidth_tooltip.show_tooltip)
    nbd_kde_bandwidth_label.bind("<Leave>", nbd_kde_bandwidth_tooltip.hide_tooltip)
    nbd_kde_bandwidth_input=ttk.Entry(master=networkburstdetectionparameters)
    nbd_kde_bandwidth_input.grid(row=2, column=1, padx=10, pady=10, sticky='w')

    # Set up all the output parameters
    outputparameters=ttk.LabelFrame(parameterframe.interior, text='Output/data manipulation', style="Custom.TLabelframe")
    outputparameters.grid(row=2, column=0, padx=10, pady=10, sticky='nsew')

    def removeinactivefunc():
        if removeinactivevar.get():
            activitythinput.configure(state='enabled')
        else:
            activitythinput.configure(state='disabled')

    # Remove inactive electrodes
    removeinactivelabel=ttk.Label(master=outputparameters, text="Remove inactive electrodes:", font=(font,10))
    removeinactivelabel.grid(row=0, column=0, padx=10, pady=10, sticky='w')
    removeinactivetooltip = Tooltip(removeinactivelabel, 'Should inactive electrodes be used when calculating well features?')
    removeinactivelabel.bind("<Enter>", removeinactivetooltip.show_tooltip)
    removeinactivelabel.bind("<Leave>", removeinactivetooltip.hide_tooltip)
    removeinactivevar=IntVar()
    removeinactiveinput=ttk.Checkbutton(outputparameters, onvalue=True, offvalue=False, variable=removeinactivevar, command=removeinactivefunc)
    removeinactiveinput.grid(row=0, column=1, padx=10, pady=10, sticky='w')

    # Setup the activity threshold
    activitythlabel=ttk.Label(master=outputparameters, text="Activity threshold:", font=(font,10))
    activitythlabel.grid(row=1, column=0, padx=10, pady=10, sticky='w')
    activitythtooltip = Tooltip(activitythlabel, 'Define the minimal activity a channel has to have, to be used in calculating well features.\nValue should be given in hertz, so a value of 0.1 would mean\nany channel with less that 1 spike per 10 seconds will be removed')
    activitythlabel.bind("<Enter>", activitythtooltip.show_tooltip)
    activitythlabel.bind("<Leave>", activitythtooltip.hide_tooltip)
    activitythinput=ttk.Entry(master=outputparameters)
    activitythinput.grid(row=1, column=1, padx=10, pady=10, sticky='w')

    def splitdatafunc():
        if splitdatavar.get():
            splitdatapartsinput.configure(state='enabled')
        else:
            splitdatapartsinput.configure(state='disabled')

    # Split data
    splitdatalabel=ttk.Label(master=outputparameters, text="Split data:", font=(font,10))
    splitdatalabel.grid(row=2, column=0, padx=10, pady=10, sticky='w')
    splitdatatooltip = Tooltip(splitdatalabel, 'Should the data be split up in shorter parts?\nSplitting data might be useful to turn one long measurement into several smaller measurements')
    splitdatalabel.bind("<Enter>", splitdatatooltip.show_tooltip)
    splitdatalabel.bind("<Leave>", splitdatatooltip.hide_tooltip)
    splitdatavar=IntVar()
    splitdatainput=ttk.Checkbutton(outputparameters, onvalue=True, offvalue=False, variable=splitdatavar, command=splitdatafunc)
    splitdatainput.grid(row=2, column=1, padx=10, pady=10, sticky='w')

    # Setup the activity threshold
    splitdatapartslabel=ttk.Label(master=outputparameters, text="Parts:", font=(font,10))
    splitdatapartslabel.grid(row=3, column=0, padx=10, pady=10, sticky='w')
    splitdatapartstooltip = Tooltip(splitdatapartslabel, 'In how many smaller parts should the data be split?')
    splitdatapartslabel.bind("<Enter>", splitdatapartstooltip.show_tooltip)
    splitdatapartslabel.bind("<Leave>", splitdatapartstooltip.hide_tooltip)
    splitdatapartsinput=ttk.Entry(master=outputparameters)
    splitdatapartsinput.grid(row=3, column=1, padx=10, pady=10, sticky='w')

    # Use multiprocessing
    multiprocessingframe=ttk.Labelframe(master=parameterframe.interior, text="Other", style="Custom.TLabelframe")
    multiprocessingframe.grid(row=2, column=1, padx=10, pady=10, sticky='nesw')
    multiprocessinglabel=ttk.Label(master=multiprocessingframe, text="Use multiprocessing:", font=(font,10))
    multiprocessinglabel.grid(row=0, column=0, padx=10, pady=10, sticky='e')
    multiprocessingtooltip = Tooltip(multiprocessinglabel, 'Using multiprocessing means the electrodes will be analyzed in parallel, generally speeding up\nthe analysis. Multiprocessing might not work properly if the device\n you\'re using does not have sufficient RAM/CPU-cores')
    multiprocessinglabel.bind("<Enter>", multiprocessingtooltip.show_tooltip)
    multiprocessinglabel.bind("<Leave>", multiprocessingtooltip.hide_tooltip)
    multiprocessingvar=IntVar()
    multiprocessinginput=ttk.Checkbutton(multiprocessingframe, onvalue=True, offvalue=False, variable=multiprocessingvar)
    multiprocessinginput.grid(row=0, column=1, padx=10, pady=10, sticky='w')

    def set_default_parameters():
        # reset dropdown parameters
        dropdown_var.set("DMP_noisebased")
        networkth_var.set('Yen')
        # reset checkboxes
        removeinactivevar.set(True)
        splitdatavar.set(False)
        multiprocessingvar.set(False)
        splitdatapartsinput.configure(state='enabled')
        # Update other parameter availability
        if removeinactivevar.get():
            activitythinput.configure(state='enabled')
        else:
            activitythinput.configure(state='disabled')

        validation_method = dropdown_var.get()
        if validation_method=='DMP_noisebased':
            exittimeinput.configure(state="enabled")
            maxheightinput.configure(state="enabled")
            amplitudedropinput.configure(state="enabled")
        else:
            exittimeinput.configure(state="disabled")
            maxheightinput.configure(state="disabled")
            amplitudedropinput.configure(state="disabled")

        # reset the entry labels
        list=[lowcutoffinput, highcutoffinput, orderinput, thresholdportioninput, stdevmultiplierinput, RMSmultiplierinput, refractoryperiodinput, exittimeinput,
            amplitudedropinput, maxheightinput, minspikesinput, defaultthinput, maxisiinput, isikdebwinput, smallernbinput,
            minchannelsinput, activitythinput, splitdatapartsinput, nbd_kde_bandwidth_input]
        defaults=[200, 3500, 2, 0.1, 5, 5, 0.001, 0.001, 5, 2, 5, 100, 1000, 1, 10, 0.5, 0.1, 10, 0.05]
        counter=0
        for parameter in list:
            parameter.delete(0, END)
            parameter.insert(0, defaults[counter])
            counter+=1
        hertzinput.delete(0, END)
        electrodeamntinput.delete(0, END)
        if splitdatavar.get():
            splitdatapartsinput.configure(state='enabled')
        else:
            splitdatapartsinput.configure(state='disabled')
        
    set_default_parameters()

    default_parameters=ttk.Button(master=parameterframe.interior, text="Restore default parameters", command=set_default_parameters)
    default_parameters.grid(row=3, column=1, padx=10, pady=(10,20), sticky='nsew')
    parameterframe.interior.rowconfigure(3, minsize=125)

    def import_parameters():
        parametersfile = filedialog.askopenfilename(filetypes=[("Parameter file", "*.json")])
        parameters=json.load(open(parametersfile))
        # Enable all the entries so we can set the correct values
        entries=[activitythinput, exittimeinput, maxheightinput, amplitudedropinput, splitdatapartsinput]
        for entry in entries:
            entry.configure(state="enabled")        
        # Set all the parameters to the values of the imported file
        hertzinput.delete(0, END)
        hertzinput.insert(0, parameters["sampling rate"])
        electrodeamntinput.delete(0, END)
        electrodeamntinput.insert(0, parameters["electrode amount"])
        lowcutoffinput.delete(0, END)
        lowcutoffinput.insert(0, parameters["low cutoff"])
        highcutoffinput.delete(0, END)
        highcutoffinput.insert(0, parameters["high cutoff"])
        orderinput.delete(0, END)
        orderinput.insert(0, parameters["order"])
        thresholdportioninput.delete(0, END)
        thresholdportioninput.insert(0, parameters["threshold portion"])
        stdevmultiplierinput.delete(0, END)
        stdevmultiplierinput.insert(0, parameters["standard deviation multiplier"])
        RMSmultiplierinput.delete(0, END)
        RMSmultiplierinput.insert(0, parameters["rms multiplier"])
        refractoryperiodinput.delete(0, END)
        refractoryperiodinput.insert(0, parameters["refractory period"])
        dropdown_var.set(parameters["spike validation method"])
        exittimeinput.delete(0, END)
        exittimeinput.insert(0, parameters["exit time"])
        amplitudedropinput.delete(0, END)
        amplitudedropinput.insert(0, parameters["drop amplitude"])
        maxheightinput.delete(0, END)
        maxheightinput.insert(0, parameters["max drop amount"])
        minspikesinput.delete(0, END)
        minspikesinput.insert(0, parameters["minimal amount of spikes"])
        defaultthinput.delete(0, END)
        defaultthinput.insert(0, parameters["default interval threshold"])
        maxisiinput.delete(0, END)
        maxisiinput.insert(0, parameters["max interval threshold"])
        isikdebwinput.delete(0, END)
        isikdebwinput.insert(0, parameters["KDE bandwidth"])
        smallernbinput.delete(0, END)
        smallernbinput.insert(0, parameters["smaller neighbours"])
        minchannelsinput.delete(0, END)
        minchannelsinput.insert(0, parameters["min channels"])
        networkth_var.set(parameters["thresholding method"])
        removeinactivevar.set(bool(parameters["remove inactive electrodes"]))
        activitythinput.delete(0, END)
        activitythinput.insert(0, parameters["activity threshold"])
        splitdatavar.set(bool(parameters["split data"]))
        splitdatapartsinput.delete(0, END)
        splitdatapartsinput.insert(0, parameters["parts"])
        multiprocessingvar.set(bool(parameters["use multiprocessing"]))
        nbd_kde_bandwidth_input.delete(0, END)
        nbd_kde_bandwidth_input.insert(0, parameters["nbd_kde_bandwidth"])
            
        # Update other parameter availability
        if removeinactivevar.get():
            activitythinput.configure(state='enabled')
        else:
            activitythinput.configure(state='disabled')

        validation_method = dropdown_var.get()
        if validation_method=='DMP_noisebased':
            exittimeinput.configure(state="enabled")
            maxheightinput.configure(state="enabled")
            amplitudedropinput.configure(state="enabled")
        else:
            exittimeinput.configure(state="disabled")
            maxheightinput.configure(state="disabled")
            amplitudedropinput.configure(state="disabled")

        if splitdatavar.get():
            splitdatapartsinput.configure(state='enabled')
        else:
            splitdatapartsinput.configure(state='disabled')

    default_parameters=ttk.Button(master=parameterframe.interior, text="Import parameters", command=import_parameters)
    default_parameters.grid(row=3, column=2, padx=10, pady=(10,20), sticky='nsew')


    def data_analysis():
        def close_progressbar():
            # Communicate to the progressbar that the analysis has crashed
            progressfile=f'{os.path.split(filename)[0]}/progress.npy'
            np.save(progressfile, ['crashed'])
        if multiprocessingvar.get():
            try:
                analyse_wells(fileadress=filename, wells=wells, hertz=hertz, validation_method=validation_method, low_cutoff=low_cutoff, high_cutoff=high_cutoff, order=order, spikeduration=spikeduration,
                        exit_time_s=exit_time_s, electrode_amnt=electrode_amnt, bd_kde_bandwidth=kde_bandwidth, smallerneighbours=smallerneighbours, minspikes_burst=minspikes_burst,
                        max_threshold=max_threshold, default_threshold=default_threshold, max_drop_amount=max_drop_amount, amplitude_drop_sd=amplitude_drop_sd,
                        stdevmultiplier=stdevmultiplier, RMSmultiplier=RMSmultiplier, min_channels=min_channels, threshold_method=threshold_method, nbd_kde_bandwidth=global_nbd_kde_bandwidth, activity_threshold=activity_threshold,
                        threshold_portion=threshold_portion, remove_inactive_electrodes=remove_inactive_electrodes, cut_data_bool=cut_data_bool, parts=parts, use_multiprocessing=True)
            except Exception as error:
                traceback.print_exc()
                tk.messagebox.showerror(title='Error', message='Something went wrong with analyzing the data, please check if all the parameters are set correctly\nAlternatively, try analyzing the data with multiprocessing turned off')
                close_progressbar()
        else:
            try:
                analyse_wells(fileadress=filename, wells=wells, hertz=hertz, validation_method=validation_method, low_cutoff=low_cutoff, high_cutoff=high_cutoff, order=order, spikeduration=spikeduration,
                            exit_time_s=exit_time_s, electrode_amnt=electrode_amnt, bd_kde_bandwidth=kde_bandwidth, smallerneighbours=smallerneighbours, minspikes_burst=minspikes_burst,
                            max_threshold=max_threshold, default_threshold=default_threshold, max_drop_amount=max_drop_amount, amplitude_drop_sd=amplitude_drop_sd,
                            stdevmultiplier=stdevmultiplier, RMSmultiplier=RMSmultiplier, min_channels=min_channels, threshold_method=threshold_method, nbd_kde_bandwidth=global_nbd_kde_bandwidth, activity_threshold=activity_threshold,
                            threshold_portion=threshold_portion, remove_inactive_electrodes=remove_inactive_electrodes, cut_data_bool=cut_data_bool, parts=parts, use_multiprocessing=False)
            except Exception as error:
                traceback.print_exc()
                tk.messagebox.showerror(title='Error', message='Something went wrong with analyzing the data, please check if all the parameters are set correctly')
                close_progressbar()

    def progress():
        # Read out the progressbar from the file
        start=time.time()
        path=os.path.split(filename)[0]
        progressfile=f"{path}/progress.npy"
        popup=tk.Toplevel(root)
        popup.title('Progress')
        try:
            popup.iconbitmap(os.path.join(application_path))
        except Exception as error:
            print(error)
        progressinfo=ttk.Label(master=popup, text='The MEA data is being analyzed, please wait for the program to finish')
        progressinfo.grid(row=0, column=0, pady=10, padx=20)
        progressbarlength=1000
        progressbar=ttk.Progressbar(master=popup, length=progressbarlength, mode='determinate', maximum=progressbarlength)
        progressbar.grid(row=1, column=0, pady=10, padx=20)
        progressbar.step(0)
        info=ttk.Label(master=popup, text='')
        info.grid(row=2, column=0, pady=10, padx=20)
        while True:
            currenttime=time.time()
            elapsed=round(currenttime-start,1)
            try:
                progress=np.load(progressfile)
            except:
                progress=['starting']
            if progress[0]=='crashed':
                popup.destroy()
                os.remove(progressfile)
                return
            if progress[0]=='done':
                break
            elif progress[0]=='starting':
                info.configure(text=f'Loading raw data, time elapsed: {elapsed} seconds')
            else:
                currentprogress=(progress[0]/progress[1])*progressbarlength
                progressbar.configure(value=currentprogress)
                info.configure(text=f"Analyzing data, channel: {progress[0]}/{progress[1]}, time elapsed: {elapsed} seconds")
            time.sleep(0.01)
        popup.destroy()
        os.remove(progressfile)
        currenttime=time.time()
        elapsed=round(currenttime-start,2)
        finishedpopup=tk.Toplevel(root)
        finishedpopup.title('Finished')
        try:
            finishedpopup.iconbitmap(os.path.join(application_path))
        except Exception as error:
            print(error)
        finishedtext=ttk.Label(master=finishedpopup, text=f'The data has been analyzed. It took {elapsed} seconds')
        finishedtext.grid(column=0, row=0, padx=10, pady=10)
        
    # Start the analysis
    def go_to_analysisframe():
        # Check if the user has selected a file
        if filename == "":
            tk.messagebox.showerror(title='Error', message='No file has been selected yet, please select a file and set parameters')
            return
        # Start two seperate threads for analyzing the data and reading out the progress. This makes sure the main gui stays responsive
        data_thread=threading.Thread(target=data_analysis)
        data_thread.start()
        progress_thread=threading.Thread(target=progress)
        progress_thread.start()

    def go_to_resultfileframe():
        resultfileframe.pack(fill='both', expand=True)
        main_frame.pack_forget()
        

    analysisframe=ttk.Frame(root)
    analysisframe.pack(fill='both', expand=True)
    analysisframe.pack_forget()

    btn_parameter=ttk.Button(master=main_frame, text="Start\nAnalysis", style="bigbutton.TButton", command=go_to_analysisframe)
    btn_parameter.grid(row=0, column=1, padx=10, pady=10, sticky='nesw', ipadx=25, ipady=25)

    # Select files to view results
    resultfileframe=ttk.Frame(root)
    resultfileframe.pack(fill='both', expand=True)
    resultfileframe.pack_forget()

    results_btn=ttk.Button(master=main_frame, text="View\nResults", style="bigbutton.TButton", command=go_to_resultfileframe)
    results_btn.grid(row=1, column=0, padx=10, pady=10, sticky='nesw', ipadx=25, ipady=25)

    def results_file_to_main_func():
        main_frame.pack(fill='both', expand=True)
        resultfileframe.pack_forget()

    # Results file selection
    datalocation=ttk.LabelFrame(resultfileframe, text='Data location', style="Custom.TLabelframe")
    datalocation.grid(row=0, column=0, padx=10, pady=10, sticky='nsew')

    resultsfolder=''    # default value
    def openfolder():
        global resultsfolder
        resultsfolder = filedialog.askdirectory()
        if resultsfolder != '':
            btn_selectfolder.configure(text=resultsfolder)

    # Open location of results
    folderlabel=ttk.Label(master=datalocation, text="Folder:", font=(font,10))
    folderlabel.grid(row=0, column=0, padx=10, pady=10)
    btn_selectfolder=ttk.Button(master=datalocation, text="Select a folder", command=openfolder)
    btn_selectfolder.grid(row=0, column=1, padx=10, pady=10)

    def openrawfile():
        global filename
        selectedfile = filedialog.askopenfilename(filetypes=[("MEA data", "*.h5")])
        if len(selectedfile)!=0:
            filename=selectedfile
            btn_selectrawfile.configure(text=os.path.split(filename)[1])

    # Open location of raw data
    rawfilelabel=ttk.Label(master=datalocation, text="File:", font=(font,10))
    rawfilelabel.grid(row=1, column=0, padx=10, pady=10)
    btn_selectrawfile=ttk.Button(master=datalocation, text="Choose a file", command=openrawfile)
    btn_selectrawfile.grid(row=1, column=1, padx=10, pady=10)

    # View the results
    resultsframe=ttk.Frame(root)
    resultsframe.pack(fill='both', expand=True)
    resultsframe.pack_forget()

    # Results notebook
    results_nb=ttk.Notebook(master=resultsframe)
    results_nb.pack(fill='both', expand=True)

    electrode_frame=ttk.Frame(master=results_nb)
    electrode_frame.pack(fill='both', expand=True)

    well_frame=ttk.Frame(master=results_nb)
    well_frame.pack(fill='both', expand=True)

    output_frame=ttk.Frame(master=results_nb)
    output_frame.pack(fill='both', expand=True)

    results_nb.add(electrode_frame, text='Single electrode view')
    results_nb.add(well_frame, text='Whole well view')
    results_nb.add(output_frame, text="Results")

    '''Build the output_frame tab'''
    table_frame=ttk.Frame(master=output_frame)
    table_frame.grid(row=0, column=0, sticky='nesw', padx=25, pady=25)
    output_frame.grid_columnconfigure(0, weight=1)
    output_frame.grid_rowconfigure(0, weight=1)

    output_frame.grid_rowconfigure(1, weight=1)

    global raw_data

    class treeview_table(ttk.Frame):
        def __init__(self, master, file_path):
            super().__init__(master)  # Initialize the parent class
            self.tableframe=ttk.Frame(self)
            self.tableframe.grid(row=0, column=0, sticky='nesw')
            self.rowconfigure(0, weight=1)
            self.columnconfigure(0, weight=1)
            self.tableframe.rowconfigure(0, weight=1)
            self.tableframe.columnconfigure(0, weight=1)

            self.button_frame=ttk.Frame(self)
            self.button_frame.grid(row=1, column=0, sticky='ew', pady=5)
            self.groupmanager=ttk.Frame(self.button_frame)
            self.groupmanager.grid(row=0, column=1, sticky='ew')
            self.group_frame=ttk.Frame(self)
            self.group_frame.grid(row=3, column=0, sticky='ew')

            self.file_path = file_path
            df = pd.read_csv(file_path)
            self.columns=list(df.columns)

            vert_scrollbar = ttk.Scrollbar(self.tableframe, orient="vertical")
            vert_scrollbar.grid(row=0, column=1, sticky='nesw')

            hor_scrollbar = ttk.Scrollbar(self.tableframe, orient="horizontal")
            hor_scrollbar.grid(row=1, column=0, sticky='nesw')

            self.tree = ttk.Treeview(self.tableframe, yscrollcommand=vert_scrollbar.set, xscrollcommand=hor_scrollbar.set)
            self.tree.grid(row=0, column=0, sticky='nesw')

            vert_scrollbar.config(command=self.tree.yview)
            hor_scrollbar.config(command=self.tree.xview)
            # Define the columns
            self.tree["columns"] = self.columns
            width=100
            # Format the columns
            self.tree.column("#0", width=0, stretch=tk.NO)
            self.tree.heading("#0", text="")
            for feature in self.columns:
                self.tree.column(feature, anchor=tk.W, width=width)
                self.tree.heading(feature, text=feature)

            # Add data
            data = df.values.tolist()

            for item in data:
                self.tree.insert(parent='', index=tk.END, values=list(np.round(item,2)))
            self.currenttag=''
            def item_selected(event):
                for selected_item in self.tree.selection():
                    # If the user tries to set the same tag again, deselect the row
                    if len(self.tree.item(selected_item)['tags'])!=0 and self.tree.item(selected_item)['tags'][0]==self.currenttag:
                        self.tree.item(selected_item, tags=())
                    else:
                        self.tree.item(selected_item, tags=self.currenttag)
                self.deselect_all()
            self.tree.bind('<<TreeviewSelect>>', item_selected)

            clearbutton=ttk.Button(self.button_frame, text='Clear all groups', command=self.clear_selection)
            clearbutton.grid(row=0, column=0, sticky='w')

            self.groups=[]
            self.buttons=[]

            add_group_button=ttk.Button(self.groupmanager, text='Add group:', command=self.add_group)
            add_group_button.grid(row=0, column=1, sticky='w')

            self.add_group_entry=ttk.Entry(self.groupmanager)
            self.add_group_entry.grid(row=0, column=2, sticky='w', padx=(0,10))

            self.colors= [
                "#FF5733",  # Vivid Red
                "#3357FF",  # Vivid Blue
                "#FF33A6",  # Vivid Pink
                "#33FFF0",  # Vivid Cyan
                "#FFD700",  # Vivid Yellow
                "#8A2BE2",  # Blue Violet
                "#FF4500",  # Orange Red
                "#32CD32",  # Lime Green
                "#9400D3",  # Dark Violet
                "#33FF57",  # Vivid Green
            ]

            # Create styles
            self.style = ttk.Style()
            for i, color in enumerate(self.colors):
                style_name = f"Custom{i}.TButton"
                self.style.configure(style_name, foreground=color)
                self.style.map(style_name, foreground=[('active', color)])

        def get_selection(self):
            dict={}
            for group in self.groups:
                temp=[]
                for item in self.tree.get_children():
                    if len(self.tree.item(item)['tags'])!=0:
                        if str(self.tree.item(item)['tags'][0])==group:
                            temp.append(self.tree.item(item)['values'])
                dict[group]=temp
            return dict, self.columns, self.colors

        def set_tag(self, tag):
            self.currenttag=str(tag)

        def clear_selection(self):
            for item in self.tree.get_children():
                self.tree.item(item, tags=())

        def add_group(self):
            new_group=str(self.add_group_entry.get())
            if new_group != '' and new_group not in self.groups and len(self.groups)<10:
                self.groups.append(new_group)
            self.update_buttons()

        def update_buttons(self):
            # Remove buttons
            for button in self.buttons:
                button.destroy()
            for group in range(len(self.groups)):
                # Create tag
                self.tree.tag_configure(str(self.groups[group]), background=self.colors[group])
                # Create buttons
                button=ttk.Button(self.group_frame, text=self.groups[group], command=partial(self.set_tag, str(self.groups[group])), style=f"Custom{group}.TButton")
                button.grid(row=0, column=group, sticky='w')
                self.buttons.append(button)
        
        def deselect_all(self):
            for item in self.tree.selection():
                self.tree.selection_remove(item)

    class graphs(ttk.Frame):
        def __init__(self, master, table):
            super().__init__(master)  # Initialize the parent class
            self.table=table
            self.buttonframe=ttk.Frame(self)
            self.buttonframe.grid(row=0, column=0, sticky='nesw')
            self.treeframe=ttk.Frame(self)
            self.treeframe.grid(row=1, column=0, sticky='nesw')
            
            self.update_graph_button=ttk.Button(self.buttonframe, text='Update table', command=self.update_table)
            self.update_graph_button.grid(row=0, column=0, sticky='new', pady=10)
            self.update_plot_button=ttk.Button(self.buttonframe, text='Create graph', command=self.update_plot)
            self.update_plot_button.grid(row=0, column=1, sticky='new', pady=10)
            self.disclaimer_label=ttk.Label(self.buttonframe, text="\nThe calculated p-value is merely to indicate where significant differences might be located\nIt should not be used to draw conclusions as it does not take into account the origin/source or distribution of the data")
            self.disclaimer_label.grid(row=0, column=2, pady=10, padx=10, sticky='new')

            self.rowconfigure(1, weight=1)
            self.columnconfigure(0, weight=1)
            self.plot_values={}

            # Configure the treeview
            vert_scrollbar = ttk.Scrollbar(self.treeframe, orient="vertical")
            vert_scrollbar.grid(row=0, column=1, sticky='nesw')
            self.tree = ttk.Treeview(self.treeframe, yscrollcommand=vert_scrollbar.set)
            self.tree.grid(row=0, column=0, sticky='nesw')
            vert_scrollbar.config(command=self.tree.yview)
            # Define the columns
            headers=["Feature", "p-value"]
            self.tree["columns"] = headers
            width=200
            # # Format the columns
            self.tree.column("#0", width=0, stretch=tk.NO)
            self.tree.heading("#0", text="")
            for header in headers:
                self.tree.column(header, anchor=tk.W, width=width)
                self.tree.heading(header, text=header)

            # colors for the boxplots
            self.palette=[]       
            # Create the fig and canvas
            self.figure = Figure()
            self.canvas = FigureCanvasTkAgg(self.figure, master=self.treeframe)
            self.canvas.get_tk_widget().grid(row=0, column=2, sticky='nesw')
            self.canvas.draw()

            self.treeframe.columnconfigure(2, weight=1) 
            self.treeframe.rowconfigure(0, weight=1)

            self.groups=[]

            if theme=='dark':
                self.bgcolor='#1c1c1c'
                self.axiscolour='#ecf3fa'
            else:
                self.bgcolor='#fafafa'
                self.axiscolour='#221c1c'
            self.figure.set_facecolor(self.bgcolor)


        def update_table(self):
            data, columns, colors=self.table.get_selection()
            # Retrieve the colors of the groups that are not empty
            temp_colors=[]
            temp_groups=[]
            keys=list(data.keys())
            for i in range(len(keys)):
                if len(data[keys[i]])!=0:
                    temp_colors.append(colors[i])
                    temp_groups.append(keys[i])
            self.palette=temp_colors
            self.groups=temp_groups
            # Remove entries where the list is empty
            data = {key: value for key, value in data.items() if value}
            if len(data)==0:
                tk.messagebox.showerror(title='Error', message='No rows selected')
                return
            if len(data)==1:
                tk.messagebox.showerror(title='Error', message='Not enough groups to compare, select at least 2 groups')
                return
            p_values=[]
            keys=list(data.keys())
            test_title="P_value"
            # If we have 2 groups, do t-tests for all features to make a ranking
            if len(data)==2:
                test_title="P_value (t-test)"
                group1=np.array(data[keys[0]]).astype(float)
                group2=np.array(data[keys[1]]).astype(float)
                # Skip the first column because this is the well number
                for i in range(1,len(columns)):
                    t_stat, p_value = stats.ttest_ind(group1[:,i], group2[:,i])
                    p_values.append(p_value)
                    self.plot_values[columns[i]]=[group1[:,i], group2[:,i]]
            # If we have more than 2 groups, perform an anova
            if len(data)>2:
                test_title="P_value (ANOVA)"
                for i in range(1,len(columns)):
                    anova_data=[]
                    for key in keys:
                        anova_data.append(np.array(data[key]).astype(float)[:,i])
                    t_stat, p_value=stats.f_oneway(*anova_data)
                    p_values.append(p_value)
                    self.plot_values[columns[i]]=anova_data
            self.update_graph_button.configure(text="Update table")
            
            headers=["Feature", test_title]
            self.tree["columns"] = headers
            for header in headers:
                self.tree.heading(header, text=header)

            # Clear the treeview
            for item in self.tree.get_children():
                self.tree.delete(item)
            # Insert the new values
            for i in range(len(p_values)):
                self.tree.insert(parent='', index=tk.END, values=[columns[i+1], p_values[i]])

            self.tree.tag_configure('selected', background="#bb86fc")

            def item_selected(event):
                # Either select or deselect the items
                for selected_item in self.tree.selection():
                    if len(self.tree.item(selected_item)['tags'])!=0:
                        self.tree.item(selected_item, tags=())
                    else:
                        self.tree.item(selected_item, tags='selected')
                for item in self.tree.selection():
                    self.tree.selection_remove(item)
            self.tree.bind('<<TreeviewSelect>>', item_selected)

        def update_plot(self):
            labels=[]
            for item in self.tree.get_children():
                if len(self.tree.item(item)['tags'])!=0:
                    labels.append(self.tree.item(item)['values'][0])
            if len(labels)>0:
                self.update_plot_button.configure(text='Update graph')

                # First remove all previous axes
                while self.figure.axes:
                    self.figure.delaxes(self.figure.axes[0])
                # Create multiple different plots
                self.axes = self.figure.subplots(1, len(labels))
                for i in range(len(labels)):
                    if len(labels)==1:
                        ax=self.axes
                    else:
                        ax=self.axes.flatten()[i]
                    data_df=[]
                    for array in range(len(self.plot_values[labels[i]])):
                        df = pd.DataFrame({'Group': self.groups[array], 'Value': self.plot_values[labels[i]][array]})
                        data_df.append(df)
                    data_df = pd.concat(data_df, ignore_index=True)
                    sns.boxplot(x='Group', y='Value', data=data_df, palette=self.palette, ax=ax)
                    sns.stripplot(x='Group', y='Value', data=data_df, jitter=True, palette=[self.axiscolour]*len(self.palette), ax=ax)
                    #ax.set_xticks(ticks=range(len(self.groups)), labels=self.groups)
                    ax.set_title(f"{labels[i]}")

                # Set the correct colours
                for ax in self.figure.axes:
                    ax.set_facecolor(self.bgcolor)
                    # Change the other colours
                    ax.xaxis.label.set_color(self.axiscolour)
                    ax.yaxis.label.set_color(self.axiscolour)
                    for side in ['top', 'bottom', 'left', 'right']:
                        ax.spines[side].set_color(self.axiscolour)
                    ax.tick_params(axis='x', colors=self.axiscolour)
                    ax.tick_params(axis='y', colors=self.axiscolour)
                    ax.set_title(label=ax.get_title(),color=self.axiscolour)
                self.canvas.draw()

    def load_table(file_path):
        global feature_filepath
        results_window=tk.Toplevel(master=output_frame)
        try:
            results_window.iconbitmap(os.path.join(application_path))
        except Exception as error:
            print(error)
        results_window.title("Results")
        results_window.geometry("500x500") # Set dimensions in pixels 
        table=treeview_table(results_window, feature_filepath)
        table.grid(row=0,column=0, sticky='nesw')
        graph=graphs(results_window, table)
        graph.grid(row=1, column=0, sticky='nesw')
        results_window.columnconfigure(0, weight=1)
        results_window.rowconfigure(0, weight=2)
        results_window.rowconfigure(1, weight=1)

    # Add buttons for selecting which results to load
    feature_filepath=''
    def get_feature_filepath():
        global feature_filepath
        temp=filedialog.askopenfilename(filetypes=[("Feature file", "*.csv")])
        if temp != '':
            feature_filepath=temp
        choose_featurefile_button.configure(text=feature_filepath)

    choose_featurefile_button=ttk.Button(master=output_frame, text="Choose a Features.csv file to analyse", command=get_feature_filepath)
    choose_featurefile_button.grid(row=0, column=0, pady=10, padx=10, sticky='n')
    load_table_button=ttk.Button(master=output_frame, text='Load table', command=lambda: load_table(feature_filepath))
    load_table_button.grid(row=1, column=0, pady=10, padx=10, sticky='n')

    global selected_well
    selected_well = 1
    selected_electrode = 1

    view_results=ttk.Button(master=resultfileframe, text="View results", compound=LEFT, command=lambda: threading.Thread(target=view_results_func).start())
    view_results.grid(row=1, column=0, padx=10, pady=10, sticky='nsew')

    results_file_to_main=ttk.Button(master=resultfileframe, text="Back", command=results_file_to_main_func)
    results_file_to_main.grid(row=2, column=0, padx=10, pady=10, sticky='nsew')

    # Single electrode frame
    choose_well=ttk.LabelFrame(master=electrode_frame, text='Select a well', style="Custom.TLabelframe")
    choose_well.grid(column=0, row=0, sticky='nw', padx=10)

    # Choose electrode
    choose_electrode=ttk.LabelFrame(master=electrode_frame, text='Select an electrode', style="Custom.TLabelframe")
    choose_electrode.grid(column=1, row=0, sticky='nw', padx=10)

    def well_button_pressed(well):
        global selected_well
        selected_well=well

    # Function to calculate the initial grid
    def calculate_optimal_grid(num_wells):
        min_difference = num_wells
        optimal_width = num_wells
        optimal_height = 1
        
        for width in range(1, int(math.sqrt(num_wells)) + 1):
            if num_wells % width == 0:
                height = num_wells // width
                difference = abs(width - height)
                if difference < min_difference:
                    min_difference = difference
                    optimal_width = width
                    optimal_height = height
        
        return int(optimal_width), int(optimal_height)

    def create_wellbuttons(master, wells, button_command):
        ywells, xwells=calculate_optimal_grid(wells)
        wellbuttons=[]
        i=1

        for y in range(ywells):
            for x in range(xwells):
                well_btn=ttk.Button(master=master, text=i, command=partial(button_command, i), style="wellbutton.TButton")
                well_btn.grid(row=y, column=x, sticky='nesw', ipadx=25, ipady=25)
                wellbuttons.append(well_btn)
                i+=1
        return wellbuttons

    '''This function takes the amount of electrodes the MEA has, and creates a correctly
    sized grid with buttons for it'''
    def create_electrodebuttons(master, electrode_amnt, button_command):
        # Two presets for a 12 and 16 electrode MEA
        if electrode_amnt==12:
            electrode_layout=np.array([ [False, True, True, False],
                                        [True, True, True, True],
                                        [True, True, True, True],
                                        [False, True, True, False]])
        elif electrode_amnt==16:    
            electrode_layout=np.array([ [True, True, True, True],
                                        [True, True, True, True],
                                        [True, True, True, True],
                                        [True, True, True, True]])
        # If the MEA has a different amount of electrodes than 12 or 16, we calculate the grid with the following function
        else:
            height, width = calculate_optimal_grid(electrode_amnt)
            electrode_layout = np.ones((height, width), dtype=bool)
        i = 1
        electrodebuttons=[]

        # Loop over the grid
        for x in range(electrode_layout.shape[0]):
            for y in range(electrode_layout.shape[1]):
                if electrode_layout[x,y]:
                    # And create a button for each of the electrodes
                    electrode_btn=ttk.Button(master=master, text=i, command=partial(button_command, i), style="mediumbutton.TButton")
                    electrode_btn.grid(row=x, column=y, sticky='nesw', ipadx=25, ipady=25)
                    electrodebuttons.append(electrode_btn)
                    i+=1
        return electrodebuttons

    def plot_network_bursts(master, parameters, well):
        global resultsfolder
        fig=network_burst_detection(outputpath=resultsfolder, wells=[well], electrode_amnt=parameters["electrode amount"], measurements=parameters["measurements"], hertz=parameters["sampling rate"], min_channels=parameters["min channels"], threshold_method=parameters["thresholding method"], bandwidth=parameters["nbd_kde_bandwidth"], plot_electrodes=True, savedata=False, save_figures=False)
        if theme=='dark':
            bgcolor='#1c1c1c'
            axiscolour='#ecf3fa'
        else:
            bgcolor='#fafafa'
            axiscolour='#221c1c'
        
        fig.set_facecolor(bgcolor)
        for ax in fig.axes:
            ax.set_facecolor(bgcolor)
            # Change the other colours
            ax.xaxis.label.set_color(axiscolour)
            ax.yaxis.label.set_color(axiscolour)
            for side in ['top', 'bottom', 'left', 'right']:
                ax.spines[side].set_color(axiscolour)
            ax.tick_params(axis='x', colors=axiscolour)
            ax.tick_params(axis='y', colors=axiscolour)
            ax.set_title(label=ax.get_title(),color=axiscolour)

        plot_canvas = FigureCanvasTkAgg(fig, master=master)  
        plot_canvas.get_tk_widget().grid(row=0, column=0, sticky='nsew')
        toolbarframe=ttk.Frame(master=master)
        toolbarframe.grid(row=1, column=0, sticky='s')
        toolbar = NavigationToolbar2Tk(plot_canvas, toolbarframe)
        toolbar.update()
        plot_canvas.draw()

    def plot_electrode_activity(master, parameters, well, bandwidth):
        global resultsfolder
        fig=well_electrodes_kde(outputpath=resultsfolder, well=well, electrode_amnt=parameters["electrode amount"], measurements=parameters["measurements"], hertz=parameters["sampling rate"], bandwidth=bandwidth)
        if theme=='dark':
            bgcolor='#1c1c1c'
            axiscolour='#ecf3fa'
        else:
            bgcolor='#fafafa'
            axiscolour='#221c1c'
        fig.set_facecolor(bgcolor)
        for ax in fig.axes:
            ax.set_facecolor(bgcolor)
            # Change the other colours
            ax.xaxis.label.set_color(axiscolour)
            ax.yaxis.label.set_color(axiscolour)
            for side in ['top', 'bottom', 'left', 'right']:
                ax.spines[side].set_color(axiscolour)
            ax.tick_params(axis='x', colors=axiscolour)
            ax.tick_params(axis='y', colors=axiscolour)
            ax.set_title(label=ax.get_title(),color=axiscolour)

        plot_canvas = FigureCanvasTkAgg(fig, master=master)  
        plot_canvas.get_tk_widget().grid(row=0, column=0, sticky='nsew')
        toolbarframe=ttk.Frame(master=master)
        toolbarframe.grid(row=1, column=0)
        toolbar = NavigationToolbar2Tk(plot_canvas, toolbarframe)
        toolbar.update()
        plot_canvas.draw()

    def network_visualization(well):
        global resultsfolder
        # Open a new window and setup the notebook
        main_well_window=tk.Toplevel(master=well_frame)
        main_well_window.title(f'Well: {well}')
        try:
            main_well_window.iconbitmap(os.path.join(application_path))
        except Exception as error:
            print(error)

        well_nb=ttk.Notebook(master=main_well_window)
        well_nb.pack(fill='both', expand=True)
        
        network_burst_detection=ttk.Frame(master=well_nb)
        network_burst_detection.pack(expand=True, fill='both')

        electrode_activity=ttk.Frame(master=well_nb)
        electrode_activity.pack(expand=True, fill='both')

        well_nb.add(network_burst_detection, text='Network burst detection')
        well_nb.add(electrode_activity, text='Electrode activity')

        '''Network burst detection'''
        network_burst_plot_frame=ttk.Frame(master=network_burst_detection)
        network_burst_plot_frame.grid(row=0, column=0, sticky='nesw')
        network_burst_plot_frame.grid_columnconfigure(0, weight=1)
        network_burst_plot_frame.grid_rowconfigure(0, weight=1)

        # Load the parameters
        parameters=open(f"{resultsfolder}/parameters.json")
        parameters=json.load(parameters)

        # Initial plot
        plot_network_bursts(master=network_burst_plot_frame, parameters=parameters, well=well)

        network_burst_detection.grid_columnconfigure(0, weight=1)
        network_burst_detection.grid_rowconfigure(0, weight=3)

        # Frame for the NB settings
        nb_settings_frame=ttk.Labelframe(master=network_burst_detection, text="Network burst detection settings", style="Custom.TLabelframe")
        nb_settings_frame.grid(row=1, column=0, padx=10, pady=10)

        # NB settings disclaimer
        nbd_disclaimer_label=ttk.Label(master=nb_settings_frame, text="Warning: altering these settings will have no effect on the output of the application\nor further analysis steps such as feature calculation.\nThese settings are here purely to visualize how these parameters could alter the analysis pipeline")
        nbd_disclaimer_label.grid(row=0, column=2, rowspan=2, pady=10, padx=10, sticky='w')

        # NB settings
        min_channels_nb_label=ttk.Label(master=nb_settings_frame, text="Min channels (%)").grid(row=0, column=0, sticky='w', padx=10, pady=10)
        min_channels_nb_entry=ttk.Entry(master=nb_settings_frame)
        min_channels_nb_entry.grid(row=0, column=1, sticky='w', padx=10, pady=10)
        
        th_method_nb_var = tk.StringVar(nb_settings_frame)
        nwthoptions_nb = ['Yen', 'Otsu', 'Li', 'Isodata', 'Mean', 'Minimum', 'Triangle']
        th_method_nb = ttk.Label(master=nb_settings_frame, text="Thresholding method:")
        th_method_nb.grid(row=1, column=0, padx=10, pady=10, sticky='nesw')
        th_method_dropdown_nb = ttk.OptionMenu(nb_settings_frame, th_method_nb_var, nwthoptions_nb[0], *nwthoptions_nb)
        th_method_dropdown_nb.grid(row=1, column=1, padx=10, pady=10, sticky='w')

        nbd_kde_bandwidth_nb_label=ttk.Label(master=nb_settings_frame, text="KDE Bandwidth:").grid(row=2, column=0, sticky='w', padx=10, pady=10)
        nbd_kde_bandwidth_nb_entry=ttk.Entry(master=nb_settings_frame)
        nbd_kde_bandwidth_nb_entry.grid(row=2, column=1, sticky='w', padx=10, pady=10)

        def nb_default_values():
            th_method_nb_var.set(parameters["thresholding method"])
            min_channels_nb_entry.delete(0,END)
            min_channels_nb_entry.insert(0,parameters["min channels"])
            nbd_kde_bandwidth_nb_entry.delete(0, END)
            nbd_kde_bandwidth_nb_entry.insert(0, parameters["nbd_kde_bandwidth"])
            

        nb_default_values()

        def nb_update_plot():
            temp_parameters=copy.deepcopy(parameters)
            temp_parameters["min channels"]=float(min_channels_nb_entry.get())
            temp_parameters["thresholding method"]=str(th_method_nb_var.get())
            temp_parameters["nbd_kde_bandwidth"]=float(nbd_kde_bandwidth_nb_entry.get())

            plot_network_bursts(master=network_burst_plot_frame, parameters=temp_parameters, well=well)

        def nb_reset():
            nb_default_values()
            nb_update_plot()

        nb_update_plot_button=ttk.Button(master=nb_settings_frame, text="Update plot", command=nb_update_plot)
        nb_update_plot_button.grid(row=3, column=0, sticky='nesw', padx=10, pady=10)

        nb_reset_button=ttk.Button(master=nb_settings_frame, text="Reset", command=nb_reset)
        nb_reset_button.grid(row=3, column=1, sticky='nesw', padx=10, pady=10)

        '''Electrode activity'''
        electrode_activity_plot_frame=ttk.Frame(master=electrode_activity)
        electrode_activity_plot_frame.grid(row=0, column=0, sticky='nsew')
        electrode_activity_plot_frame.grid_columnconfigure(0, weight=1)
        electrode_activity_plot_frame.grid_rowconfigure(0, weight=1)
        electrode_activity.grid_columnconfigure(0, weight=1)
        electrode_activity.grid_rowconfigure(0, weight=1)

        def_bw_value=0.1
        plot_electrode_activity(master=electrode_activity_plot_frame, parameters=parameters, well=well, bandwidth=def_bw_value)

        electrode_activity_settings=ttk.Frame(master=electrode_activity)
        electrode_activity_settings.grid(row=1, column=0)

        el_act_bw_label=ttk.Label(master=electrode_activity_settings, text="KDE bandwidth")
        el_act_bw_label.grid(row=0, column=0, sticky='nesw', padx=10, pady=10)
        el_act_bw_entry=ttk.Entry(master=electrode_activity_settings)
        el_act_bw_entry.grid(row=0, column=1, sticky='nesw', pady=10, padx=10)
        el_act_bw_entry.insert(0, def_bw_value)

        def update_electrode_activity():
            plot_electrode_activity(master=electrode_activity_plot_frame, parameters=parameters, well=well, bandwidth=float(el_act_bw_entry.get()))
        
        el_act_update_plot=ttk.Button(master=electrode_activity_settings, text='Update plot', command=update_electrode_activity)
        el_act_update_plot.grid(row=1, column=0, sticky='nesw', padx=10, pady=10)

        def reset_electrode_activity():
            el_act_bw_entry.delete(0, END)
            el_act_bw_entry.insert(0, def_bw_value)
            update_electrode_activity()

        el_act_reset=ttk.Button(master=electrode_activity_settings, text='Reset', command=reset_electrode_activity)
        el_act_reset.grid(row=1, column=1, sticky='nesw', padx=10, pady=10)


    choose_well_nw=ttk.LabelFrame(master=well_frame, text='Select a well', style="Custom.TLabelframe")
    choose_well_nw.grid(column=0, row=0, sticky='nesw', padx=10)

    '''This function handles all the preparation necessary for properly viewing the analysed data'''
    def view_results_func():
        global resultsfolder

        # Load in the correct files
        global raw_data
        # Check if the correct files have been selected
        if os.path.exists(f"{resultsfolder}/burst_values") and os.path.exists(f"{resultsfolder}/spike_values") and os.path.exists(f"{resultsfolder}/network_data"):
            # Load the parameters
            try:
                parameters=open(f"{resultsfolder}/parameters.json")
                parameters=json.load(parameters)
            except Exception as error:
                traceback.print_exc()
                tk.messagebox.showerror(title='Error', message='Could not load in the results, please make sure you have selected the correct folder')
                return
            # Load the raw data
            try:
                loaddatapopup=tk.Toplevel(resultfileframe)
                loaddatapopup.title('Progress')
                try:
                    loaddatapopup.iconbitmap(os.path.join(application_path))
                except Exception as error:
                    traceback.print_exc()
                progressinfo=ttk.Label(master=loaddatapopup, text='Loading in the raw data...')
                progressinfo.grid(row=0, column=0, pady=10, padx=20)
                raw_data=openHDF5(filename)
            except Exception as error:
                traceback.print_exc()
                tk.messagebox.showerror(title='Error', message='Could not load in the raw data, please make sure you have selected the correct file')
                loaddatapopup.destroy()
                return
        else:
            tk.messagebox.showerror(title='Error', message='Could not load in the results, please make sure you have selected the correct folder')
            return
        global feature_filepath
        feature_filepath=f"{resultsfolder}/Features.csv"
        choose_featurefile_button.configure(text=feature_filepath)
        electrode_wellbuttons=create_wellbuttons(choose_well, raw_data.shape[0]/parameters["electrode amount"], well_button_pressed)
        whole_wellbuttons=create_wellbuttons(choose_well_nw, raw_data.shape[0]/parameters["electrode amount"], network_visualization)
        electrode_buttons=create_electrodebuttons(choose_electrode, parameters["electrode amount"], electrode_button_pressed)
        loaddatapopup.destroy()
        resultsframe.pack(fill='both', expand=True)
        resultfileframe.pack_forget()

    '''Button for going back to the begin frame'''
    # def results_buttonframe_to_main_func():
    #     raw_data=[]
    #     main_frame.pack(fill='both', expand=True)
    #     resultsframe.pack_forget()

    # resultsframe_to_mainmenu_button=ttk.Button(master=electrode_frame, text="Return to main menu", command=results_buttonframe_to_main_func)
    # resultsframe_to_mainmenu_button.grid(row=1, column=0, pady=10, padx=10, sticky='nsew', columnspan=2)    


    def plot_single_electrode(master, parameters, electrode_nr, plot_rectangle):
        electrode_data=butter_bandpass_filter(raw_data[electrode_nr], lowcut=parameters["low cutoff"], highcut=parameters["high cutoff"], fs=parameters["sampling rate"], order=parameters["order"])
        threshold=fast_threshold(electrode_data, hertz=parameters["sampling rate"], stdevmultiplier=parameters["standard deviation multiplier"], RMSmultiplier=parameters["rms multiplier"], threshold_portion=parameters["threshold portion"])
        fig=spike_validation(data=electrode_data, electrode=electrode_nr, threshold=threshold, hertz=parameters["sampling rate"], spikeduration=parameters["refractory period"], exit_time_s=parameters["exit time"], amplitude_drop=parameters["drop amplitude"], plot_electrodes=True, electrode_amnt=parameters["electrode amount"], max_drop_amount=parameters["max drop amount"], outputpath='', savedata=False, plot_rectangles=plot_rectangle)
        # Check which colorscheme we have to use
        if theme=='dark':
            bgcolor='#1c1c1c'
            axiscolour='#ecf3fa'
        else:
            bgcolor='#fafafa'
            axiscolour='#221c1c'
        # Set the plot background
        fig.set_facecolor(bgcolor)
        ax=fig.axes[0]
        ax.set_facecolor(bgcolor)

        # Change the other colours
        ax.xaxis.label.set_color(axiscolour)
        ax.yaxis.label.set_color(axiscolour)
        for side in ['top', 'bottom', 'left', 'right']:
            ax.spines[side].set_color(axiscolour)
        ax.tick_params(axis='x', colors=axiscolour)
        ax.tick_params(axis='y', colors=axiscolour)
        ax.set_title(label=ax.get_title(),color=axiscolour)

        plot_canvas = FigureCanvasTkAgg(fig, master=master)  
        plot_canvas.get_tk_widget().grid(row=0, column=0, sticky='nsew')
        toolbarframe=ttk.Frame(master=master)
        toolbarframe.grid(row=1, column=0, sticky='s')
        toolbar = NavigationToolbar2Tk(plot_canvas, toolbarframe)
        toolbar.update()
        plot_canvas.draw()

    def plot_burst_detection(master, parameters, electrode_nr):
        global resultsfolder
        electrode_data=butter_bandpass_filter(raw_data[electrode_nr], lowcut=parameters["low cutoff"], highcut=parameters["high cutoff"], fs=parameters["sampling rate"], order=parameters["order"])
        KDE_fig, burst_fig = burst_detection(data=electrode_data, electrode=electrode_nr, electrode_amnt=parameters["electrode amount"], hertz=parameters["sampling rate"], kde_bandwidth=parameters["KDE bandwidth"], smallerneighbours=parameters["smaller neighbours"], minspikes_burst=parameters["minimal amount of spikes"], max_threshold=parameters["max interval threshold"], default_threshold=parameters["default interval threshold"], outputpath=resultsfolder, plot_electrodes=True, savedata=False)
        
        # Change the colours of the graph
        if theme=='dark':
            bgcolor='#1c1c1c'
            axiscolour='#ecf3fa'
        else:
            bgcolor='#fafafa'
            axiscolour='#221c1c'
        for fig in [KDE_fig, burst_fig]:
            fig.set_facecolor(bgcolor)
            ax=fig.axes[0]
            ax.set_facecolor(bgcolor)
            # Change the other colours
            ax.xaxis.label.set_color(axiscolour)
            ax.yaxis.label.set_color(axiscolour)
            for side in ['top', 'bottom', 'left', 'right']:
                ax.spines[side].set_color(axiscolour)
            ax.tick_params(axis='x', colors=axiscolour)
            ax.tick_params(axis='y', colors=axiscolour)
            ax.set_title(label=ax.get_title(),color=axiscolour)
        # Plot the raw burst plot
        burst_canvas = FigureCanvasTkAgg(burst_fig, master=master)  
        burst_canvas.get_tk_widget().grid(row=0, column=0, sticky='nsew')
        bursttoolbarframe=ttk.Frame(master=master)
        bursttoolbarframe.grid(row=1, column=0, sticky='s', columnspan=2)
        bursttoolbar = NavigationToolbar2Tk(burst_canvas, bursttoolbarframe)
        bursttoolbar.update()
        burst_canvas.draw()

        # Plot the KDE plot
        KDE_canvas = FigureCanvasTkAgg(KDE_fig, master=master)  
        KDE_canvas.get_tk_widget().grid(row=0, column=1, sticky='nsew')
        # KDEtoolbarframe=ttk.Frame(master=master)
        # KDEtoolbarframe.grid(row=1, column=1, sticky='s')
        # KDEtoolbar = NavigationToolbar2Tk(KDE_canvas, KDEtoolbarframe)
        # KDEtoolbar.update()

        master.grid_columnconfigure(0, weight=3)
        master.grid_columnconfigure(1, weight=1)
        master.grid_rowconfigure(0, weight=1)
        KDE_canvas.draw()

    def electrode_button_pressed(electrode):
        global selected_electrode
        global selected_well
        selected_electrode=electrode

        # Open a new window
        main_electrode_window=tk.Toplevel(master=electrode_frame)
        main_electrode_window.title(f'Well: {selected_well}, electrode: {selected_electrode}')
        try:
            main_electrode_window.iconbitmap(os.path.join(application_path))
        except Exception as error:
            print(error)

        electrode_nb=ttk.Notebook(master=main_electrode_window)
        electrode_nb.pack(fill='both', expand=True)
        
        electrode_window=ttk.Frame(master=electrode_nb)
        electrode_window.pack(expand=True, fill='both')

        burst_frame=ttk.Frame(master=electrode_nb)
        burst_frame.pack(fill='both', expand=True)

        electrode_nb.add(electrode_window, text='Spike detection')
        electrode_nb.add(burst_frame, text='Burst detection')

        # Load the parameters
        global resultsfolder
        parameters=open(f"{resultsfolder}/parameters.json")
        parameters=json.load(parameters)
        electrode_nr=(selected_well-1)*parameters["electrode amount"]+selected_electrode-1

        # Button to create/update the plot
        def update_plot():
            # Get the new parameters from the entry widgets
            # Create a temporary dict we will give to the plot electrode function
            temp_parameters=copy.deepcopy(parameters)
            temp_parameters['low cutoff']=int(lowcut_ew_entry.get())
            temp_parameters['high cutoff']=int(highcut_ew_entry.get())
            temp_parameters['order']=int(order_ew_entry.get())
            temp_parameters['standard deviation multiplier']=float(stdevmultiplier_ew_entry.get())
            temp_parameters['rms multiplier']=float(RMSmultiplier_ew_entry.get())
            temp_parameters['threshold portion']=float(thpn_ew_entry.get())
            temp_parameters['refractory period']=float(rfpd_ew_entry.get())
            if str(validation_method_var.get())=='none':
                temp_parameters['drop amplitude']=0
            else:
                temp_parameters['exit time']=float(exittime_ew_entry.get())
                temp_parameters['drop amplitude']=float(dropamplitude_ew_entry.get())
                temp_parameters['max drop amount']=float(maxdrop_ew_entry.get())

            # Plot the electrode with the new parameters
            plot_single_electrode(electrode_window, temp_parameters, electrode_nr, plot_rectangle.get())

        # Set the weights
        electrode_window.grid_columnconfigure(0, weight=1)
        electrode_window.grid_rowconfigure(0, weight=3)
        electrode_window.grid_rowconfigure(2, weight=1)

        # Create a frame for the settings
        electrode_settings_frame=ttk.Frame(master=electrode_window)
        electrode_settings_frame.grid(row=2, column=0)

        # Bandpass options
        bp_options_ew_frame = ttk.LabelFrame(master=electrode_settings_frame, text='Bandpass Filter', style="Custom.TLabelframe")
        bp_options_ew_frame.grid(row=0, column=0, pady=10, padx=10, sticky='nsew')

        lowcut_label=ttk.Label(master=bp_options_ew_frame, text='Low cutoff', font=(font,10)).grid(row=0, column=0, pady=10, padx=10, sticky='w')
        lowcut_ew_entry=ttk.Entry(master=bp_options_ew_frame)
        lowcut_ew_entry.grid(row=0, column=1, pady=10, padx=10, sticky='w')
        highcut_label=ttk.Label(master=bp_options_ew_frame, text='High cutoff', font=(font,10)).grid(row=1, column=0, sticky='w', pady=10, padx=10)
        highcut_ew_entry=ttk.Entry(master=bp_options_ew_frame)
        highcut_ew_entry.grid(row=1, column=1, pady=10, padx=10, sticky='w')
        order_label=ttk.Label(master=bp_options_ew_frame, text='Order', font=(font,10)).grid(row=2, column=0, sticky='w', pady=10, padx=10)
        order_ew_entry=ttk.Entry(master=bp_options_ew_frame)
        order_ew_entry.grid(row=2, column=1, pady=10, padx=10, sticky='w')

        # Threshold options
        th_options_ew_frame = ttk.LabelFrame(master=electrode_settings_frame, text='Threshold', style="Custom.TLabelframe")
        th_options_ew_frame.grid(row=0, column=1, pady=10, padx=10, sticky='nesw')
        
        stdevmultiplier_label=ttk.Label(master=th_options_ew_frame, text='Standard deviation multiplier', font=(font,10)).grid(row=0, column=0, pady=10, padx=10, sticky='w')
        stdevmultiplier_ew_entry=ttk.Entry(master=th_options_ew_frame)
        stdevmultiplier_ew_entry.grid(row=0, column=1, pady=10, padx=10, sticky='w')
        RMSmultiplier_label=ttk.Label(master=th_options_ew_frame, text='RMS multiplier', font=(font,10)).grid(row=1, column=0, sticky='w', pady=10, padx=10)
        RMSmultiplier_ew_entry=ttk.Entry(master=th_options_ew_frame)
        RMSmultiplier_ew_entry.grid(row=1, column=1, pady=10, padx=10, sticky='w')
        thpn_label=ttk.Label(master=th_options_ew_frame, text='Threshold portion', font=(font,10)).grid(row=2, column=0, sticky='w', pady=10, padx=10)
        thpn_ew_entry=ttk.Entry(master=th_options_ew_frame)
        thpn_ew_entry.grid(row=2, column=1, pady=10, padx=10, sticky='w')

        # Spike validation options
        val_options_ew_frame = ttk.LabelFrame(master=electrode_settings_frame, text='Spike validation', style="Custom.TLabelframe")
        val_options_ew_frame.grid(row=0, column=2, pady=10, padx=10)

        def set_states():
            validation_method = validation_method_var.get()
            if validation_method=='DMP_noisebased':
                exittime_ew_entry.configure(state="enabled")
                maxdrop_ew_entry.configure(state="enabled")
                dropamplitude_ew_entry.configure(state="enabled")
                plot_rectangle_ew_entry.configure(state="enabled")
            else:
                exittime_ew_entry.configure(state="disabled")
                maxdrop_ew_entry.configure(state="disabled")
                dropamplitude_ew_entry.configure(state="disabled")
                plot_rectangle.set(False)
                plot_rectangle_ew_entry.configure(state="disabled")
        def ew_option_selected(event):
            set_states()
        
        
        validation_method_var = tk.StringVar(validationparameters)
        validation_options = ['DMP_noisebased', 'none']
        validation_method_label = ttk.Label(master=val_options_ew_frame, text="Spike validation method:", font=(font,10))
        validation_method_label.grid(row=0, column=0, padx=10, pady=10, sticky='w')
        validation_method_entry = ttk.OptionMenu(val_options_ew_frame, validation_method_var, validation_options[0], *options, command=ew_option_selected)
        validation_method_entry.grid(row=0, column=1, padx=10, pady=10, sticky='nesw')

        rfpd_label=ttk.Label(master=val_options_ew_frame, text='Refractory period', font=(font,10)).grid(row=1, column=0, pady=10, padx=10, sticky='w')
        rfpd_ew_entry=ttk.Entry(master=val_options_ew_frame)
        rfpd_ew_entry.grid(row=1, column=1, pady=10, padx=10, sticky='w')
        exittime_label=ttk.Label(master=val_options_ew_frame, text='Exit time', font=(font,10)).grid(row=2, column=0, pady=10, padx=10, sticky='w')
        exittime_ew_entry=ttk.Entry(master=val_options_ew_frame)
        exittime_ew_entry.grid(row=2, column=1, pady=10, padx=10, sticky='w')
        dropamplitude_label=ttk.Label(master=val_options_ew_frame, text='Drop amplitude', font=(font,10)).grid(row=0, column=2, pady=10, padx=10, sticky='w')
        dropamplitude_ew_entry=ttk.Entry(master=val_options_ew_frame)
        dropamplitude_ew_entry.grid(row=0, column=3, pady=10, padx=10, sticky='w')
        maxdrop_label=ttk.Label(master=val_options_ew_frame, text='Max drop amount', font=(font,10)).grid(row=1, column=2, pady=10, padx=10, sticky='w')
        maxdrop_ew_entry=ttk.Entry(master=val_options_ew_frame)
        maxdrop_ew_entry.grid(row=1, column=3, pady=10, padx=10, sticky='w')
        plot_rectangle_label=ttk.Label(master=val_options_ew_frame, text='Plot validation', font=(font,10))
        plot_rectangle_label.grid(row=2, column=2, pady=10, padx=10, sticky='w')
        plot_rectangle_tooltip = Tooltip(plot_rectangle_label, 'Display the rectangles that have been used to validate the spikes\nWarning: Plotting these is computationally expensive and might take a while')
        plot_rectangle_label.bind("<Enter>", plot_rectangle_tooltip.show_tooltip)
        plot_rectangle_label.bind("<Leave>", plot_rectangle_tooltip.hide_tooltip)
        plot_rectangle=BooleanVar()
        plot_rectangle_ew_entry=ttk.Checkbutton(master=val_options_ew_frame, variable=plot_rectangle)
        plot_rectangle_ew_entry.grid(row=2, column=3, pady=10, padx=10, sticky='w')

        # Insert the default values from the json in the entries
        def default_values():
            # Bandpass
            lowcut_ew_entry.delete(0,END)
            lowcut_ew_entry.insert(0,parameters["low cutoff"])
            highcut_ew_entry.delete(0,END)
            highcut_ew_entry.insert(0,parameters["high cutoff"])
            order_ew_entry.delete(0,END)
            order_ew_entry.insert(0,parameters["order"])
            # Threshold
            stdevmultiplier_ew_entry.delete(0,END)
            stdevmultiplier_ew_entry.insert(0,parameters["standard deviation multiplier"])
            RMSmultiplier_ew_entry.delete(0,END)
            RMSmultiplier_ew_entry.insert(0,parameters["rms multiplier"])
            thpn_ew_entry.delete(0,END)
            thpn_ew_entry.insert(0,parameters["threshold portion"])
            # Spike validation
            rfpd_ew_entry.delete(0,END)
            rfpd_ew_entry.insert(0,parameters["refractory period"])
            exittime_ew_entry.delete(0,END)
            exittime_ew_entry.insert(0,parameters["exit time"])
            dropamplitude_ew_entry.delete(0,END)
            dropamplitude_ew_entry.insert(0,parameters["drop amplitude"])
            maxdrop_ew_entry.delete(0,END)
            maxdrop_ew_entry.insert(0,parameters["max drop amount"])
            plot_rectangle.set(False)
            validation_method_var.set(parameters['spike validation method'])
            set_states()

        def reset():
            # First, enable all the possibly disabled entries so we can alter the values
            exittime_ew_entry.configure(state="enabled")
            maxdrop_ew_entry.configure(state="enabled")
            dropamplitude_ew_entry.configure(state="enabled")
            plot_rectangle_ew_entry.configure(state="enabled")
            # Insert the new values
            default_values()
            # Update the availability of certain entries
            set_states()
            # Update the plot
            update_plot()

        # Set default values and create initial plot
        reset()
        
        update_plot_button=ttk.Button(master=electrode_settings_frame, text='Update plot', command=update_plot)
        update_plot_button.grid(row=1, column=0, pady=10, padx=10, sticky='nsew')

        reset_button = ttk.Button(master=electrode_settings_frame, text='Reset', command=reset)
        reset_button.grid(row=1, column=1, pady=10, padx=10, sticky='nsew')

        disclaimerlabel=ttk.Label(master=electrode_settings_frame, text='Warning: altering these settings will have no effect on the output of the application,\nor further analysis steps such as burst detection.\nThese settings are here purely to visualize how these parameters could alter the analysis pipeline', font=(font,10), anchor='e')
        disclaimerlabel.grid(row=1, column=2, pady=10, padx=10, sticky='nsew')

        # Now do the same for the burst window

        # Create the initial burst plot
        burstplotsframe=ttk.Frame(master=burst_frame)
        burstplotsframe.grid(row=0, column=0, sticky='nesw')

        burst_frame.grid_columnconfigure(0, weight=1)
        burst_frame.grid_rowconfigure(0, weight=3)

        burstsettingsframe=ttk.Labelframe(master=burst_frame, text='Burst detection settings', style="Custom.TLabelframe")
        burstsettingsframe.grid(row=1, column=0)
        burst_frame.grid_rowconfigure(1, weight=1)

        # Burst detection settings
        minspikes_bw_label=ttk.Label(master=burstsettingsframe, text='Minimal amount of spikes', font=(font,10)).grid(row=0, column=0, pady=10, padx=10, sticky='w')
        minspikes_bw_entry=ttk.Entry(master=burstsettingsframe)
        minspikes_bw_entry.grid(row=0, column=1, pady=10, padx=10, sticky='w')
        def_iv_bw_label=ttk.Label(master=burstsettingsframe, text='Default interval threshold', font=(font,10)).grid(row=1, column=0, pady=10, padx=10, sticky='w')
        def_iv_bw_entry=ttk.Entry(master=burstsettingsframe)
        def_iv_bw_entry.grid(row=1, column=1, pady=10, padx=10, sticky='w')
        max_iv_bw_label=ttk.Label(master=burstsettingsframe, text='Max interval threshold', font=(font,10)).grid(row=2, column=0, pady=10, padx=10, sticky='w')
        max_iv_bw_entry=ttk.Entry(master=burstsettingsframe)
        max_iv_bw_entry.grid(row=2, column=1, pady=10, padx=10, sticky='w')
        kde_bw_bw_label=ttk.Label(master=burstsettingsframe, text='KDE bandwidth', font=(font,10)).grid(row=0, column=2, pady=10, padx=10, sticky='w')
        kde_bw_bw_entry=ttk.Entry(master=burstsettingsframe)
        kde_bw_bw_entry.grid(row=0, column=3, pady=10, padx=10, sticky='w')
        smaller_nb_bw_label=ttk.Label(master=burstsettingsframe, text='Smaller neighbours', font=(font,10)).grid(row=1, column=2, pady=10, padx=10, sticky='w')
        smaller_nb_bw_entry=ttk.Entry(master=burstsettingsframe)
        smaller_nb_bw_entry.grid(row=1, column=3, pady=10, padx=10, sticky='w')

        def update_burst_plot():
            # Get the new parameters from the entry widgets
            temp_parameters=copy.deepcopy(parameters)
            temp_parameters["minimal amount of spikes"]=int(minspikes_bw_entry.get())
            temp_parameters["default interval threshold"]=float(def_iv_bw_entry.get())
            temp_parameters["max interval threshold"]=float(max_iv_bw_entry.get())
            temp_parameters["KDE bandwidth"]=float(kde_bw_bw_entry.get())
            temp_parameters["smaller neighbours"]=int(smaller_nb_bw_entry.get())

            plot_burst_detection(master=burstplotsframe, parameters=temp_parameters, electrode_nr=electrode_nr)

        def default_values_burst():
            # Reset the values to the ones in the JSON file
            minspikes_bw_entry.delete(0,END)
            minspikes_bw_entry.insert(0,parameters["minimal amount of spikes"])
            def_iv_bw_entry.delete(0,END)
            def_iv_bw_entry.insert(0,parameters["default interval threshold"])
            max_iv_bw_entry.delete(0,END)
            max_iv_bw_entry.insert(0,parameters["max interval threshold"])
            kde_bw_bw_entry.delete(0,END)
            kde_bw_bw_entry.insert(0,parameters["KDE bandwidth"])
            smaller_nb_bw_entry.delete(0,END)
            smaller_nb_bw_entry.insert(0,parameters["smaller neighbours"])

        def burst_reset():
            default_values_burst()
            update_burst_plot()

        # Set default values and create initial plot
        burst_reset()

        # Create update and reset buttons
        update_burst_plot_button=ttk.Button(master=burstsettingsframe, text='Update plot', command=update_burst_plot)
        update_burst_plot_button.grid(row=3, column=0, pady=10, padx=10, sticky='nsew', columnspan=2)

        reset_burst_button = ttk.Button(master=burstsettingsframe, text='Reset', command=burst_reset)
        reset_burst_button.grid(row=3, column=2, pady=10, padx=10, sticky='nsew', columnspan=2)

        burstdisclaimerlabel=ttk.Label(master=burstsettingsframe, text='Warning: altering these settings will have no effect on the output of the application,\nor further analysis steps such as network burst detection.\nThese settings are here purely to visualize how these parameters could alter the analysis pipeline', font=(font,10), anchor='e')
        burstdisclaimerlabel.grid(row=0, column=4, pady=10, padx=10, sticky='nsew')
    

    root.mainloop()


import matplotlib.pyplot as plt
import numpy as np
import plotly.graph_objects as go
from functools import partial
from KDEpy import FFTKDE
from matplotlib.figure import Figure
from matplotlib.gridspec import GridSpec

'''This function will calculate a 3d guassian kernel'''
def K(x, H):
    # unpack two dimensions
    x1, x2 = x
    # extract four components from the matrix inv(H)
    a, b, c, d = np.linalg.inv(H).flatten()
    
    scale = 2*np.pi*np.sqrt( np.linalg.det(H))
    return np.exp(-(a*x1**2 + d*x2**2 + (b+c)*x1*x2)/2) / scale

'''This function will insert a 3d gaussian kernel at the location of every datapoint/spike'''
def KDE(x, H, data):
    # unpack two dimensions
    x1, x2 = x
    # prepare the grid for output values
    output = np.zeros_like(x1)
    # process every sample
    for sample in data:
        output += K([x1-sample[0], x2-sample[1]], H)
    return output

'''Create a 3D view of a single well. Every spike will be represented as a 3D gaussian kernel'''
def fancyplot(outputpath,       # Where to find the spikedata
              wells,            # Which wells to generate a plot for
              electrode_amnt,   # Amount of electrodes in a well
              measurements,     # Measurements done (time*sampling rate)
              hertz,            # Sampling rate
              resolution,       # Resolution of the 3d graph - higher resolution lead to significantly more processing time
              kernel_size,      # Size of the gaussian kernel
              aspectratios,     # Ratios of the 3d plot
              colormap          # Colourmap used - can be any colourmap supported by plotly
              ):

    # Example usage:
    # fancyplot("C:/MEA_data/output_folder_of_experiment", [15], 12, 3000000, 20000, 5, 1, [0.5,0.25,0.5], "deep")

    spikepath=f'{outputpath}/spike_values'
    for well in wells:
        kdedata=[]
        for electrode in range(1,electrode_amnt+1):
            # Load in spikedata
            spikedata=np.load(f'{spikepath}/well_{well}_electrode_{electrode}_spikes.npy')
            spikedata=spikedata[:,:2]
            spikedata[:,1]=electrode
            for spike in spikedata:
                kdedata.append(spike)
        time=measurements/hertz

        # covariance matrix
        # This determines the shape of the gaussian kernel
        H = [[1, 0],
            [0, 1]]

        data = np.array(kdedata)
        # fix arguments 'H' and 'data' of KDE function for further calls
        KDE_partial = partial(KDE, H=kernel_size*np.array(H), data=data)

        # draw contour and surface plots
        func=KDE_partial

        # create a np xy grid using the dimensions of the data
        yres=int(resolution*time)
        xres=int(resolution*electrode_amnt)
        y_range = np.linspace(start=0, stop=time, num=yres)
        x_range = np.linspace(start=0, stop=electrode_amnt, num=xres)
        print(f"Creating 3D plot, resolution x: {xres} by y: {yres}")

        y_grid, X_grid = np.meshgrid(y_range, x_range)
        Z_grid = func([y_grid, X_grid])
        
        fig = go.Figure(data=[go.Surface(y=y_grid, x=X_grid, z=Z_grid, colorscale=colormap)])
        fig.update_layout(
        scene = dict(
            xaxis = dict(range=[1,electrode_amnt]),
            yaxis = dict(range=[0,time]),
            ),
        margin=dict(r=20, l=10, b=10, t=10))

        # change the aspect ratio's of the plot
        fig.update_layout(scene_aspectmode='manual',
                            title=f"Well: {well}",
                            scene_aspectratio=dict(x=(electrode_amnt/10)*aspectratios[0],
                                           y=(time/10)*aspectratios[1],
                                           z=1*aspectratios[2]))
        fig.update_layout(scene = dict(
                    xaxis_title='Electrode',
                    yaxis_title='Time in seconds',
                    zaxis_title='KDE'),)
        fig.write_html(f"{outputpath}/figures/well_{well}_3dgraph.html")

'''Plots the KDE of all electrodes (x) of a well in x amount of subplots'''
def well_electrodes_kde(outputpath,         # Where to find the spike information
                        well,               # Which well to analyse
                        electrode_amnt,     # Amount of electrodes per well
                        measurements,       # Measurements done (time*sampling rate)
                        hertz,              # Sampling frequency
                        bandwidth=1         # Bandwidth of the KDE
                        ):
    # Where to find the spike-data
    spikepath=f'{outputpath}/spike_values'
    data_time=measurements/hertz

    # Create matplotlib figure
    fig = Figure()
    
    gs = GridSpec(electrode_amnt, 1, figure=fig)
    axes=[]

    first_iteration=True
    # Loop through all electrodes
    for electrode in range(1, electrode_amnt+1):
        if first_iteration:
            ax = fig.add_subplot(gs[electrode_amnt-electrode])
            first_iteration=False
        else:
            ax = fig.add_subplot(gs[electrode_amnt-electrode], sharex=axes[0], sharey=axes[0])
        # Load in the spike data and create a KDE using KDEpy for each well
        spikedata=np.load(f'{spikepath}/well_{well}_electrode_{electrode}_spikes.npy')
        if len(spikedata[:,0])>0:
            y = FFTKDE(bw=bandwidth, kernel='gaussian').fit(spikedata[:,0]).evaluate(grid_points=np.arange(0, data_time, 0.001))
            y=np.array(y)*len(spikedata[:,0])
            x=np.arange(0, data_time, 0.001)
        else:
            x=np.arange(0, data_time, 0.001)
            y=np.zeros(len(x))
        # Plot the KDE and give the subplot a label
        ax.plot(x,y)
        ax.set_ylabel(f"E: {electrode}")
        ax.set_xlim([0, measurements/hertz])
        axes.append(ax)
    # Plot layout
    axes[-1].set_xlabel("Time in seconds")
    axes[0].set_xlim([0, measurements/hertz])
    axes[0].set_title(f"Well: {well} activity")
    
    return fig


import numpy as np
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.gridspec import GridSpec
from KDEpy import FFTKDE
from skimage import filters

'''Takes two burst values and calculates whether they overlap or not'''
def overlap(burst1, burst2):
    if burst1[1]<burst2[0] or burst2[1]<burst1[0]:
        return False
    else:
        return True

'''Network burst detection'''
def network_burst_detection(outputpath,         # Path where to retrieve and save data
                            wells,              # Which wells to analyse
                            electrode_amnt,     # Amount of electrodes in a well
                            measurements,       # Measurements done (time*sampling rate)
                            hertz,              # Sampling rate
                            min_channels,       # Percentage of channels that should be active for a network burst
                            threshold_method,   # Method to determine the automatic threshold
                            bandwidth,          # KDE bandwidth
                            plot_electrodes,    # Return plot or not
                            save_figures,       # Save plot in outputfolder or not
                            savedata=True       # Save network burst information or not
                            ):
    # Define where to retrieve information from
    spikepath=f'{outputpath}/spike_values'
    burstpath=f'{outputpath}/burst_values'

    # Calculate how many channels should be active for a network burst
    min_channels=round(min_channels*electrode_amnt)
    
    # Initialize arrays
    total_network_burst=[]
    participating_bursts=[]

    time_seconds=np.arange(0, measurements/hertz, 1/hertz)
    
    # Iterate over the wells (usually only one)
    for well in wells:
        # Create matplotlib figure
        fig = Figure(figsize=(16, 9))

        # Create a GridSpec with specified height ratios
        gs = GridSpec(4, 1, height_ratios=[3, 1, 1, 1], figure=fig)

        # Add subplots to the figure
        ax1 = fig.add_subplot(gs[0])
        ax2 = fig.add_subplot(gs[1], sharex=ax1)
        ax3 = fig.add_subplot(gs[2], sharex=ax1)
        ax4 = fig.add_subplot(gs[3], sharex=ax1)
        
        ax=[ax1, ax2, ax3, ax4]

        # Plot layout
        ax[1].set_ylabel("Active\nbursts")
        ax[0].set_xlim([0,measurements/hertz])
        ax[0].set_ylim([0,electrode_amnt+1])
        ax[0].set_yticks(np.arange(1, electrode_amnt+1, 1))
        ax[3].set_xlabel("Time in seconds")
        ax[1].set_ylim([0, electrode_amnt+1])
        ax[0].set_ylabel("Electrode")        
        ax[0].set_title(f"Well {well}")       
        ax[2].set_ylabel("Spike\nactivity\n(KDE)")
        ax[3].set_ylabel("Burst\nspike\nactivity\n(KDE)")

        # Initialize arrays
        well_spikes=[]
        burst_spikes=[]
        spikes_for_kde=np.array([])
        burst_spikes_for_kde=np.array([])

        # Iterate over all the electrodes in this well
        for electrode in range(1,electrode_amnt+1):
            # Load in spike and burst data
            spikedata=np.load(f'{spikepath}/well_{well}_electrode_{electrode}_spikes.npy')
            burstdata=np.load(f'{burstpath}/well_{well}_electrode_{electrode}_burst_spikes.npy')
            
            # Extract the spike data from the files
            well_spikes.append(spikedata[:,0])
            spikes_for_kde = np.append(spikes_for_kde, spikedata[:, 0])

            # Extract the burst data from the files
            if len(burstdata)>0:
                burst_spikes.append(burstdata[:,0])
                burst_spikes_for_kde = np.append(burst_spikes_for_kde, burstdata[:,0])
            else:
                burst_spikes.append([])
        
        '''Top plot'''
        # Plot layout
        lineoffsets1=np.arange(1,electrode_amnt+1)
        # Spikes and burst spikes in eventplot
        ax[0].eventplot(well_spikes, alpha=0.5, lineoffsets=lineoffsets1)
        ax[0].eventplot(burst_spikes, alpha=0.25, color='red', lineoffsets=lineoffsets1)

        '''2nd plot'''
        burst_freq=np.zeros(measurements)
        burst_cores_list=[]

        # Load in burst core data
        for electrode in range(1,electrode_amnt+1):
            burst_cores=np.load(f'{burstpath}/well_{well}_electrode_{electrode}_burst_cores.npy')
            burst_cores_list.append(burst_cores)
            for burstcore in burst_cores:
                burst_freq[int(burstcore[2]):int(burstcore[3])]+=1       
        ax[1].plot(time_seconds, burst_freq)

        '''3rd plot'''
        # Plot the spike frequency using kernel density estimate
        # Create the grid for the KDE
        KDE_grid=np.arange(0, measurements/hertz, (1/hertz)*10)
        if len(spikes_for_kde)>0:
            # Create the KDE using the KDEpy library
            y = FFTKDE(bw=bandwidth, kernel='gaussian').fit(spikes_for_kde).evaluate(grid_points=KDE_grid)
            x=KDE_grid
            # Normalise the kde so it ranges from 0 to 1
            y = (y - np.min(y)) / (np.max(y) - np.min(y))
            # # Plot the KDE
            ax[2].plot(x,y)

        '''Bottom plot'''
        # Check if there are any single channel bursts, if there are none, there cannot be any network bursts and we can skip the network burst detection
        if len(burst_spikes_for_kde)>0:
            # Create the KDE from only spikes that are part of a burst using the KDEpy library
            y = FFTKDE(bw=bandwidth, kernel='gaussian').fit(burst_spikes_for_kde).evaluate(grid_points=KDE_grid)
            x=KDE_grid
            # Normalise the kde so it ranges from 0 to 1
            y = (y - np.min(y)) / (np.max(y) - np.min(y))
            # # Plot the KDE
            ax[3].plot(x,y)

            # Determine the threshold using either Yen or Otsu automatic thresholding
            if threshold_method=='Yen' or threshold_method=='yen':
                threshold=filters.threshold_minimum(y, nbins=1000)
            elif threshold_method=='Otsu' or threshold_method=='otsu':
                threshold=filters.threshold_otsu(y, nbins=1000)
            elif threshold_method=='Li' or threshold_method=='li':
                threshold=filters.threshold_li(y)
            elif threshold_method=='Isodata' or threshold_method=='isodata':
                threshold=filters.threshold_isodata(y, nbins=1000)
            elif threshold_method=='Mean' or threshold_method=='mean':
                threshold=filters.threshold_mean(y)
            elif threshold_method=='Minimum' or threshold_method=='minimum':
                threshold=filters.threshold_minimum(y, nbins=1000)
            elif threshold_method=='Triangle' or threshold_method=='triangle':
                threshold=filters.threshold_triangle(y, nbins=1000)
            else:
                raise ValueError(f"\"{threshold_method}\" is not a valid thresholding method")
            
            # Plot the threshold
            ax[3].axhline(y=threshold, color='red', linestyle='-', linewidth=1)

            # Identify regions that cross the threshold
            network_burst_cores=[]
            above_threshold=y>threshold
            i=0
            # Loop through the data
            while i < len(above_threshold):
                if above_threshold[i]:
                    # A threshold crossing has been found
                    start=x[i]
                    # Keep looping while the KDE is still above the threshold
                    while above_threshold[i] and i+1<len(above_threshold):
                        i+=1
                    # When this while loop stops, the line has dropped below the threshold, now save the network burst location
                    end=x[i]
                    network_burst_cores.append([start, end])
                i+=1

            # Validate the network bursts by checking how many channels are participating
            # At the same time, remove burst cores that are too close to the beginning or end of the measurement (buffer)
            buffer=1 #second
            unvalid_burst=[]

            # Loop through the network burst cores
            for i in range(len(network_burst_cores)):
                channels_participating=0
                # Now loop through each of the channels (burst_cores_list contains all single channel bursts)
                for channel in burst_cores_list:
                    channel_participates=False
                    # Now check if there was a burst active at that channel at the time of the network burst core
                    for burst in channel:
                        if overlap(network_burst_cores[i], burst):
                            channel_participates=True
                    # If there was a burst active, increment the number of channels participating
                    if channel_participates: channels_participating+=1
                # Mark network burst cores that are too close to the start or end of the measurement
                too_close = network_burst_cores[i][0]<buffer or network_burst_cores[i][1]>(measurements/hertz)-buffer
                # Remove network bust cores that are too close to the end/start, or do not have enough participating channels
                if channels_participating<min_channels or too_close:
                    unvalid_burst.append(i)
            # Now remove all the network bursts that are not valid
            for j in sorted(unvalid_burst, reverse=True):
                del network_burst_cores[j]

            # Now calculate the outer edges of the network bursts using the single channel bursts
            # Calculate which SCBs were participating in the network bursts, and add their extremes as the outer edges of the network burst   
            participating_bursts=[]
            # Loop over the network burst cores
            for i in range(len(network_burst_cores)):
                outer_start=network_burst_cores[i][0]
                outer_end=network_burst_cores[i][1]
                # Loop through all the SCBs (burst_cores_list contains all single channel bursts)
                for j in range(len(burst_cores_list)):
                    for burst in burst_cores_list[j]:
                        # Check if the burst overlaps with the network burst core
                        if overlap(network_burst_cores[i], burst):
                            # If it starts earlier/ends later then any previous SCB, make this the start/end of the network burst
                            if burst[0]<outer_start: outer_start=burst[0]
                            if burst[1]>outer_end: outer_end=burst[1]
                            # Add the identifiers (Network_burst ID, electrode, burst ID) to the list
                            participating_bursts.append([i, j+1, int(burst[4])])
                # Add start+end to network burst cores list
                network_burst_cores[i].append(outer_start)
                network_burst_cores[i].append(outer_end)
                
                # Give each network burst a number (identifier)
                network_burst_cores[i].append(i)
            
            # Fix overlapping bursts
            overlapping_bursts=[]
            # Loop through all the bursts
            for i in range(len(network_burst_cores)-1):
                # Check if there is overlap with the next burst
                if overlap(network_burst_cores[i][2:4], network_burst_cores[i+1][2:4]):
                    overlapping_bursts.append(i)
                    overlapping_bursts.append(i+1)
            '''New method - remove outer edges of overlapping bursts'''
            if len(overlapping_bursts)>0:
                for j in overlapping_bursts:
                    network_burst_cores[j][3] = network_burst_cores[j][1]
                    network_burst_cores[j][3] = network_burst_cores[j][2]
                # Remove participating bursts with the removed ID
                participating_bursts=np.array(participating_bursts)[~np.isin(np.array(participating_bursts)[:,0],overlapping_bursts)]
            total_network_burst=(network_burst_cores)
            # Highlight the burst cores in every graph
            for graph in ax[:]:
                for network_burst in network_burst_cores:
                    temp=np.array(network_burst[:4])
                    graph.axvspan(temp[0], temp[1], facecolor='green', alpha=0.75)
                    graph.axvspan(temp[2], temp[3], facecolor='green', alpha=0.5)

        # If there were no bursts in the entire well, save nothing
        else:
            total_network_burst=[]
            participating_bursts=[]

        # Save the network burst data to a file
        if savedata:
            path = f'{outputpath}/network_data'
            np.savetxt(f'{path}/well_{well}_network_bursts.csv', total_network_burst, delimiter = ",")
            np.save(f'{path}/well_{well}_network_bursts', total_network_burst)
            np.savetxt(f'{path}/well_{well}_participating_bursts.csv', participating_bursts, delimiter = ",")
            np.save(f'{path}/well_{well}_participating_bursts', participating_bursts)
    if save_figures:
        # Save the figure
        path=f"{outputpath}/figures/well_{well}"
        fig.savefig(path, dpi=120)  # Full HD resolution (1920x1080)
    if plot_electrodes:
        return fig
    else:
        plt.cla()
        plt.clf()
        plt.close()


from scipy.signal import butter, lfilter

'''Bandpass function'''
def butter_bandpass(lowcut, highcut, fs, order=2):
    nyq = 0.5 * fs
    low = lowcut / nyq
    high = highcut / nyq
    b, a = butter(order, [low, high], btype='band')
    return b, a

def butter_bandpass_filter(data, lowcut, highcut, fs, order=2):
    b, a = butter_bandpass(lowcut, highcut, fs, order=order)
    y = lfilter(b, a, data)
    return y


import numpy as np
import h5py
import os

'''Open the HDF5 file'''
def openHDF5(adress):
    with h5py.File(adress, "r") as file_data:
        # Retrieve the data from the correct location in the hdf5 file
        dataset = file_data["Data"]["Recording_0"]["AnalogStream"]["Stream_0"]["ChannelData"]
        # Convert to numpy array
        data=dataset[:]
    return data

'''Cut the data into smaller parts'''
def cut_data(data, parts, electrodes_per_well):
    shape_before=data.shape
    total_electrodes=data.shape[0]
    total_wells=int(total_electrodes/electrodes_per_well)
    measurements=data.shape[1]
    partsize=int(measurements/parts)
    new_data=[]
    for well in range(total_wells):
        for part in range(parts):
            for electrode in range(electrodes_per_well):
                electrode_id=well*electrodes_per_well+electrode
                new_data.append(data[electrode_id, part*partsize:(part+1)*partsize])
    new_data=np.array(new_data)
    print(f"Data converted from shape: {shape_before} to shape: {new_data.shape}")
    return new_data

''' Get all of the MEA files in a folder '''
def get_files(MEA_folder):
    # Get all files from MEA folder 
    all_files = os.listdir(MEA_folder)

    MEA_files = []
    # Get all HDF5 files
    for file in all_files:
        # Convert file to right format
        file = "{0}/{1}".format(MEA_folder, file)

        # Check if file is HDF5 file
        if not file.endswith(".h5"):
            print("'{0}' is no HDF5 file!".format(file))
            continue

        # Check if HDF5 file can be opened
        try:
            h5_file = h5py.File(file, "r")
        except:
            print("'{0}' can not be opened as HDF5 file!".format(file))
            continue

        # Check if HDF5 MEA dataset object exist
        try:
            h5_file["Data"]["Recording_0"]["AnalogStream"]["Stream_0"]["ChannelData"]
        except:
            print("'{0}' has no MEA dataset object!".format(file))
            continue

        # Create list with all MEA files
        MEA_files.append(file)

    # Print all HDF5 MEA files
    print("\nList with all HDF5 MEA files:")
    for file in MEA_files:
        print(file)
    
    return MEA_files


# The init file can be left empty or you could provide a nice message as well as some variables describing your Library
Message = "Have fun with all our functions by analyzing your MEA files!"
Developers = "Developed by Joram van Beem and Jesse Antonissenfor the CureQ research consortium."


# Imports
import time
import multiprocessing
from multiprocessing.managers import SharedMemoryManager
import gc
import json
from datetime import date
from datetime import datetime
import os
from importlib.metadata import version

# External libraries
import numpy as np
import pandas as pd

# Import MEA functions
try:
    from ._bandpass import *
    from ._burst_detection import *
    from ._features import *
    from ._network_burst_detection import *
    from ._open_file import *
    from ._plotting import *
    from ._spike_validation import *
    from ._threshold import *
except:
    from _bandpass import *
    from _burst_detection import *
    from _features import *
    from _network_burst_detection import *
    from _open_file import *
    from _plotting import *
    from _spike_validation import *
    from _threshold import *

'''Analyse electrode as subprocess
This is the subproces that gets called when multiprocessing is turned on'''
def _electrode_subprocess(outputpath, memory_id, shape, _type, electrode, hertz, low_cutoff, high_cutoff, order, stdevmultiplier,
                        RMSmultiplier, threshold_portion, spikeduration, exit_time_s,
                        amplitude_drop_sd, plot_electrodes, electrode_amnt,
                        max_drop_amount, kde_bandwidth, smallerneighbours,
                        minspikes_burst, max_threshold, default_threshold,
                        validation_method, progressfile):
    # Load in the data from the shared memory block in the RAM
    existing_shm = multiprocessing.shared_memory.SharedMemory(name=memory_id, create=False)
    funcdata=np.ndarray(shape, _type, buffer=existing_shm.buf)

    # From all the data, select the electrode
    data=funcdata[electrode]

    # Filter the data
    data=butter_bandpass_filter(data, low_cutoff, high_cutoff, hertz, order)
    
    # Calculate the threshold
    threshold_value=fast_threshold(data, hertz, stdevmultiplier, RMSmultiplier, threshold_portion)
    
    # Calculate spike values
    if validation_method=="DMP_noisebased":
        spike_validation(data, electrode, threshold_value, hertz, spikeduration, exit_time_s, amplitude_drop_sd, plot_electrodes, electrode_amnt, max_drop_amount, outputpath)
    elif validation_method=='none':
        spike_validation(data, electrode, threshold_value, hertz, spikeduration, exit_time_s, 0, plot_electrodes, electrode_amnt, max_drop_amount, outputpath)
    else:
        raise ValueError(f"\"{validation_method}\" is not a valid spike validation method")
    
    # Detect the bursts
    burst_detection(data, electrode, electrode_amnt, hertz, kde_bandwidth, smallerneighbours, minspikes_burst, max_threshold, default_threshold, outputpath, plot_electrodes)

    data=None
    existing_shm.close()
    print(f"Calculated electrode: {electrode}")

'''Analyse an entire well'''
def analyse_wells(    fileadress,                                # Where is the data file stored
                      hertz,                                # What is the sampling frequency of the MEA
                      electrode_amnt,                       # The amount of electrodes per well
                      wells='all',                          # Which wells do you want to analyze
                      validation_method="DMP_noisebased",   # Which validation method do you want to use, possible: "none", "DMP_noisebased"
                      low_cutoff=200,                       # The low_cutoff for the bandpass filter
                      high_cutoff=3500,                     # The high_cutoff for the bandpass filter
                      order=2,                              # The order for the bandpass filter
                      spikeduration=0.001,                  # The amount of time only 1 spike should be registered, aka refractory period
                      exit_time_s=0.001,                    # The amount of time a spike gets to drop amplitude in the validation methods
                      bd_kde_bandwidth=1,                   # The bandwidth of the kernel density estimate
                      smallerneighbours=10,                 # The amount of smaller neighbours a peak should have before being considered as one
                      minspikes_burst=5,                    # The minimal amount of spikes a burst should have
                      max_threshold=1000,                   # The maximal ISIth2 that can be used in burst detection
                      default_threshold=100,                # The default value for ISIth1
                      max_drop_amount=2,                    # Multiplied with the spike detection threshold, will be the maximal value the box for DMP_noisebased validation can be
                      amplitude_drop_sd=5,                  # Multiplied with the SD of surrounding noise, will be the boxheight for DMP_noisebased validation
                      stdevmultiplier=5,                    # The amount of SD a value needs to be from the mean to be considered a possible spike in the threshold detection
                      RMSmultiplier=5,                      # Multiplied with the RMS of the spike-free noise, used to determine the threshold
                      min_channels=0.5,                     # Minimal % of channels that should participate in a burst
                      threshold_method='yen',               # Threshold method to decide whether activity is a network burst or not
                      nbd_kde_bandwidth=0.05,               # Bandwidth for network burst detection KDE
                      activity_threshold=0.1,               # The lowest frequency an electrode can have before being removed from the analysis
                      threshold_portion=0.1,                # How much of the electrode do you want to use to calculate the threshold. Higher number = higher runtime
                      remove_inactive_electrodes=True,      # Whether you want to remove inactive electrodes
                      cut_data_bool=False,                  # Whether you want to cut the data
                      parts=10,                             # In how many parts
                      use_multiprocessing=False             # Whether to use multiprocessing
                      ):
    analysis_time=time.time()
    plot_electrodes=False

    # Advertisement
    print(f"CureQ MEA Library - Version: {version('CureQ')}. Developed by Joram van Beem and Jesse Antonissen for the CureQ research consortium")
    print(f"Analyzing: {fileadress}")
    
    # Create a directory which will contain the output
    outputpath=os.path.splitext(fileadress)[0]#remove the file extension
    outputpath=outputpath+"_output"

    # Add current datetime
    outputpath=outputpath+f"_{date.today()}"
    now = datetime.now()
    current_time = now.strftime("%H-%M-%S")
    outputpath=outputpath+f"_{current_time}"
    os.makedirs(outputpath)

    # Create different folders for the output
    os.makedirs(f"{outputpath}/burst_values")
    os.makedirs(f"{outputpath}/spike_values")
    os.makedirs(f"{outputpath}/network_data")
    os.makedirs(f"{outputpath}/figures")

    # Create a file to commmunicate the progress with the GUI
    progressfile=f'{os.path.split(fileadress)[0]}/progress.npy'
    np.save(progressfile, ['starting'])
    
    # Call the freeze_support function to make sure multiprocessing still works properly if the algorithm is frozen
    multiprocessing.freeze_support()

    # Open the raw data
    print("Opening the data")
    data=openHDF5(fileadress)

    # Check if the electrode_amnt parameter is set properly
    if len(data)%electrode_amnt != 0:
        raise ValueError(f"The total amount of electrodes ({len(data)}) is not divisible by the number of electrodes per well ({electrode_amnt})")

    # If all wells should be analysed, generate a list of wells
    if wells=='all':
        wells=list(range(1,int(data.shape[0]/electrode_amnt)+1))

    # Should the data bet cut into smaller parts - see documentation
    if cut_data_bool:
        # Cut the data
        data=cut_data(data, parts, electrode_amnt)
        # Calculate the new well numbers
        new_wells=np.array([])
        for well in wells:
            temp=(np.arange((well-1)*parts+1, (well)*(parts)+1))
            new_wells=np.append(new_wells, temp)
        wells=new_wells.astype(int)
        wells=wells.tolist()

    # Flag for if it is the first iteration
    first_iteration=True

    # Save the parameters that have been given in a JSON file 
    parameters={
        'file adress' : fileadress,
        'wells' : wells,
        'sampling rate' : hertz,
        'electrode amount' : electrode_amnt,
        'low cutoff' : low_cutoff,
        'high cutoff' : high_cutoff,
        'order' : order,
        'threshold portion' : threshold_portion,
        'standard deviation multiplier' : stdevmultiplier,
        'rms multiplier' : RMSmultiplier,
        'refractory period' : spikeduration,
        'spike validation method' : validation_method,
        'exit time' : exit_time_s,
        'drop amplitude' : amplitude_drop_sd,
        'max drop amount' : max_drop_amount,
        'minimal amount of spikes' : minspikes_burst,
        'default interval threshold' : default_threshold,
        'max interval threshold' : max_threshold,
        'KDE bandwidth' : bd_kde_bandwidth,
        'smaller neighbours' : smallerneighbours,
        'min channels' : min_channels,
        'thresholding method' : threshold_method,
        'nbd_kde_bandwidth' : nbd_kde_bandwidth,
        'remove inactive electrodes' : remove_inactive_electrodes,
        'activity threshold' : activity_threshold,
        'split data' : cut_data_bool,
        'parts' : parts,
        'use multiprocessing' : use_multiprocessing,
        'measurements' : data.shape[1]
    }
    with open(f"{outputpath}/parameters.json", 'w') as outfile:
        json.dump(parameters, outfile)

    # With multiprocessing
    if use_multiprocessing:
        # Save the data in shared memory
        print("Loading data into shared memory")
        # Create space on the RAM
        np_size=data.nbytes
        shape=data.shape
        _type=data.dtype
        measurements=data.shape[1]
        data=None
        sharedmemory=multiprocessing.shared_memory.SharedMemory(create=True, size=np_size)
        # Communicate file size with GUI
        np.save(progressfile, [(0)*electrode_amnt, shape[0]])
        # Put data into shared memory
        data_shared=np.ndarray(shape, dtype=_type, buffer=sharedmemory.buf)
        data_shared[:]=openHDF5(fileadress)
        # Get the memory ID
        memory_id=sharedmemory.name
        # Clear up memory
        data=None

        # Start up a process for every single electrode
        print("Initializing processes")
        with multiprocessing.Pool(processes=electrode_amnt) as pool:
            # Iterate over all wells
            for well in wells:
                start=time.time()
                # Calculate which electrodes belong to this well
                electrodes=np.arange((well-1)*electrode_amnt, well*electrode_amnt)
                print(f"Analyzing well: {well}, consisting of electrodes: {electrodes}")

                # Divide the tasks to the processes
                args=[(outputpath, memory_id, shape, _type, electrode, hertz, low_cutoff, high_cutoff, order, stdevmultiplier, RMSmultiplier, threshold_portion, spikeduration, exit_time_s, amplitude_drop_sd, plot_electrodes, electrode_amnt, max_drop_amount, bd_kde_bandwidth, smallerneighbours, minspikes_burst, max_threshold, default_threshold, validation_method, progressfile) for electrode in electrodes]
                pool.starmap(_electrode_subprocess, args)

                # Calculate the network bursts
                network_burst_detection(outputpath=outputpath, wells=[well], electrode_amnt=electrode_amnt, measurements=measurements, hertz=hertz, min_channels=min_channels, threshold_method=threshold_method, bandwidth=nbd_kde_bandwidth, plot_electrodes=plot_electrodes, save_figures=True)
                print(f"Calculated network bursts well: {well}")

                # Calculate electrode and well features
                features_df=electrode_features(outputpath, well, electrode_amnt, measurements, hertz, activity_threshold, remove_inactive_electrodes)
                well_features_df=well_features(outputpath, well, electrode_amnt, measurements, hertz)
                print(f"Calculated features well: {well}")

                # If its the first iteration, create the dataframe
                if first_iteration:
                    first_iteration=False
                    output=feature_output(features_df, well_features_df, electrode_amnt)
                # If its not the first iteration, keep appending to the dataframe
                else:
                    output=pd.concat([output, feature_output(features_df, well_features_df, electrode_amnt)], axis=0)
                end=time.time()
                print(f"It took {end-start} seconds to analyse well: {well}")

                # Communicate progression with GUI
                np.save(progressfile, [(well)*electrode_amnt, shape[0]])
        # Clean up the shared memory
        sharedmemory.close()
        sharedmemory.unlink()
    # Without multiprocessing
    else:
        for well in wells:
            start=time.time()
            # Calculate which electrodes belong to this well
            electrodes=np.arange((well-1)*electrode_amnt, well*electrode_amnt)
            print(f"Analyzing well: {well}, consisting of electrodes: {electrodes}")

            # Loop through all the electrodes
            for electrode in electrodes:
                # Filter the data
                data[electrode]=butter_bandpass_filter(data[electrode], low_cutoff, high_cutoff, hertz, order)
                
                # Calculate the threshold
                threshold_value=fast_threshold(data[electrode], hertz, stdevmultiplier, RMSmultiplier, threshold_portion)
                
                # Calculate spike values
                if validation_method=="DMP_noisebased":
                    spike_validation(data[electrode], electrode, threshold_value, hertz, spikeduration, exit_time_s, amplitude_drop_sd, plot_electrodes, electrode_amnt, max_drop_amount, outputpath)
                elif validation_method=='none':
                    spike_validation(data[electrode], electrode, threshold_value, hertz, spikeduration, exit_time_s, 0, plot_electrodes, electrode_amnt, max_drop_amount, outputpath)
                else:
                    raise ValueError(f"\"{validation_method}\" is not a valid spike validation method")
                
                # Detect the bursts
                burst_detection(data[electrode], electrode, electrode_amnt, hertz, bd_kde_bandwidth, smallerneighbours, minspikes_burst, max_threshold, default_threshold, outputpath, plot_electrodes)
                # Communicate with GUI
                np.save(progressfile, [electrode+1, data.shape[0]])
                print(f"Calculated electrode: {electrode}")
            measurements=data.shape[1]

            # Detect network bursts
            network_burst_detection(outputpath=outputpath, wells=[well], electrode_amnt=electrode_amnt, measurements=measurements, hertz=hertz, min_channels=min_channels, threshold_method=threshold_method, bandwidth=nbd_kde_bandwidth, plot_electrodes=plot_electrodes, save_figures=True)
            print(f"Calculated network bursts well: {well}")

            # Calculate electrode and well features
            features_df=electrode_features(outputpath, well, electrode_amnt, measurements, hertz, activity_threshold, remove_inactive_electrodes)
            well_features_df=well_features(outputpath, well, electrode_amnt, measurements, hertz)
            print(f"Calculated features well: {well}")

            # If its the first iteration, create the dataframe
            if first_iteration:
                first_iteration=False
                output=feature_output(features_df, well_features_df, electrode_amnt)
            # If its not the first iteration, keep appending to the dataframe
            else:
                output=pd.concat([output, feature_output(features_df, well_features_df, electrode_amnt)], axis=0, ignore_index=False)
            end=time.time()
            print(f"It took {end-start} seconds to analyse well: {well}")
        
        # Free up RAM
        data=None
        del data
        gc.collect()

    # Save the output
    output.to_csv(f"{outputpath}/Features.csv", index=False)
    
    # Close the analysis
    np.save(progressfile, ['done'])
    print(f"It took {time.time()-analysis_time} seconds to analyse {fileadress}")
    print("Done")
    return output


import numpy as np

'''Threshold function - returns threshold value'''
def fast_threshold(data,                    # Raw data of a particular electrode
                   hertz,                   # Sampling frequency of MEA
                   stdevmultiplier,         # Multiplier used for determining noise
                   RMSmultiplier,           # Multiplier used for determining final threshold
                   threshold_portion        # Portion of the data that should be used for calculating the threshold (0-1), lower value takes less time.
                   ):  
        
    measurements=data.shape[0]

    # The stepsize needed for a 50ms window is calculated
    stepsize = int(0.05 * hertz)
    # Calculate how many steps there are in the measurement
    total_steps=int(measurements/stepsize)

    # Determine which parts of the data the algorithm will use for calculating the threshold
    loopvalues=range(0,total_steps)
    step=int(1/threshold_portion)
    loopvalues=loopvalues[::step]

    noise=np.array([])

    # Loop over certain position in the raw data
    for j in loopvalues:
        # Calculate the standard deviation
        std=np.std(data[j*stepsize:(j+1)*stepsize])
        # Check whether the minimum/maximal value lies outside the range of *stdevmultiplier* times the standard deviation - if this is not the case, this 50ms box can be seen as 'spike-free noise'
        if not(np.min(data[j*stepsize:(j+1)*stepsize])<(-1*stdevmultiplier*std) or np.max(data[j*stepsize:(j+1)*stepsize])>(stdevmultiplier*std)):
            # 50ms has been identified as noise and will be added to the noise paremeter
            noise = np.append(noise, data[j*stepsize:(j+1)*stepsize])

    # Calculate the RMS of all the noise
    RMS=np.sqrt(np.mean(noise**2))
    # Calculate the threshold
    threshold_value=RMSmultiplier*RMS

    return threshold_value